import os
import re
import shutil
import pickle
from datetime import date, time, datetime

import numpy as np
import pandas as pd
import pytest

from larray.tests.common import meta
from larray.tests.common import (assert_array_nan_equal, inputpath, tmp_path,
                                 needs_xlwings, needs_pytables, needs_openpyxl, must_warn)
from larray.inout.common import _supported_scalars_types
from larray import (Session, Axis, Array, Group, isnan, zeros_like, ndtest, ones_like,
                    ones, full, full_like, stack, local_arrays, global_arrays, arrays)


# avoid flake8 errors
meta = meta


def equal(o1, o2):
    if isinstance(o1, Array) or isinstance(o2, Array):
        return o1.equals(o2)
    elif isinstance(o1, Axis) or isinstance(o2, Axis):
        return o1.equals(o2)
    else:
        return o1 == o2


def assert_seq_equal(got, expected):
    assert len(got) == len(expected)
    for e1, e2 in zip(got, expected):
        assert equal(e1, e2), f"{e1} != {e2}"


a = Axis('a=a0..a2')
a2 = Axis('a=a0..a4')
a3 = Axis('a=a0..a3')
anonymous = Axis(4)
a01 = a['a0,a1'] >> 'a01'
ano01 = a['a0,a1']
b = Axis('b=0..4')
b2 = Axis('b=b0..b4')
b024 = b[[0, 2, 4]] >> 'b024'
c = 'c'
d = {}
e = ndtest([(2, 'a'), (3, 'b')])
_e = ndtest((3, 3))
f = ndtest((Axis(3), Axis(2)), dtype=float)
g = ndtest([(2, 'a'), (4, 'b')])
h = ndtest((a3, b2))
k = ndtest((3, 3))

# ########################### #
#           SESSION           #
# ########################### #


@pytest.fixture()
def session():
    return Session([('b', b), ('b024', b024), ('a', a), ('a2', a2), ('anonymous', anonymous),
                    ('a01', a01), ('ano01', ano01), ('c', c), ('d', d), ('e', e), ('g', g), ('f', f), ('h', h)])


def test_init_session(meta):
    s = Session(b, b024, a, a01, a2=a2, anonymous=anonymous, ano01=ano01, c=c, d=d, e=e, g=g, f=f, h=h)
    assert list(s.keys()) == ['b', 'b024', 'a', 'a01', 'a2', 'anonymous', 'ano01', 'c', 'd', 'e', 'g', 'f', 'h']

    # TODO: format auto-detection does not work in this case
    # s = Session('test_session_csv')
    # assert list(s.keys()) == ['e', 'f', 'g']

    # metadata
    s = Session(b, b024, a, a01, a2=a2, anonymous=anonymous, ano01=ano01, c=c, d=d, e=e, f=f, g=g, h=h, meta=meta)
    assert s.meta == meta


@needs_xlwings
def test_init_session_xlsx():
    s = Session(inputpath('demography_eurostat.xlsx'))
    assert list(s.keys()) == ['population', 'population_benelux', 'population_5_countries',
                              'births', 'deaths', 'immigration']


@needs_pytables
def test_init_session_hdf():
    s = Session(inputpath('test_session.h5'))
    assert list(s.keys()) == ['e', 'f', 'g', 'h', 'a', 'a2', 'anonymous', 'b', 'a01', 'ano01', 'b024']


def test_getitem(session):
    assert session['a'] is a
    assert session['a2'] is a2
    assert session['anonymous'] is anonymous
    assert session['b'] is b
    assert session['a01'] is a01
    assert session['ano01'] is ano01
    assert session['b024'] is b024
    assert session['c'] == 'c'
    assert session['d'] == {}
    assert equal(session['e'], e)
    assert equal(session['h'], h)


def test_getitem_list(session):
    assert list(session[[]]) == []
    assert list(session[['b', 'a']]) == [b, a]
    assert list(session[['a', 'b']]) == [a, b]
    assert list(session[['a', 'a2']]) == [a, a2]
    assert list(session[['anonymous', 'ano01']]) == [anonymous, ano01]
    assert list(session[['b024', 'a']]) == [b024, a]
    assert list(session[['e', 'a01']]) == [e, a01]
    assert list(session[['a', 'e', 'g']]) == [a, e, g]
    assert list(session[['g', 'a', 'e']]) == [g, a, e]


def test_getitem_larray(session):
    s1 = session.filter(kind=Array)
    s2 = Session({'e': e + 1, 'f': f})
    res_eq = s1[s1.element_equals(s2)]
    res_neq = s1[~(s1.element_equals(s2))]
    assert list(res_eq) == [f]
    assert list(res_neq) == [e, g, h]


def test_setitem(session):
    s = session.copy()
    s['g'] = 'g'
    assert s['g'] == 'g'


def test_getattr(session):
    assert session.a is a
    assert session.a2 is a2
    assert session.anonymous is anonymous
    assert session.b is b
    assert session.a01 is a01
    assert session.ano01 is ano01
    assert session.b024 is b024
    assert session.c == 'c'
    assert session.d == {}


def test_setattr(session):
    s = session.copy()
    s.i = 'i'
    assert s.i == 'i'


def test_add(session):
    i = Axis('i=i0..i2')
    i01 = i['i0,i1'] >> 'i01'
    session.add(i, i01, j='j')
    assert i.equals(session.i)
    assert i01 == session.i01
    assert session.j == 'j'


def test_iter(session):
    expected = [b, b024, a, a2, anonymous, a01, ano01, c, d, e, g, f, h]
    assert_seq_equal(session, expected)


def test_filter(session):
    session.ax = 'ax'
    assert_seq_equal(session.filter(), [b, b024, a, a2, anonymous, a01, ano01, 'c', {}, e, g, f, h, 'ax'])
    assert_seq_equal(session.filter('a*'), [a, a2, anonymous, a01, ano01, 'ax'])
    assert list(session.filter('a*', dict)) == []
    assert list(session.filter('a*', str)) == ['ax']
    assert list(session.filter('a*', Axis)) == [a, a2, anonymous]
    assert list(session.filter(kind=Axis)) == [b, a, a2, anonymous]
    assert list(session.filter('a01', Group)) == [a01]
    assert list(session.filter(kind=Group)) == [b024, a01, ano01]
    assert_seq_equal(session.filter(kind=Array), [e, g, f, h])
    assert list(session.filter(kind=dict)) == [{}]
    assert list(session.filter(kind=(Axis, Group))) == [b, b024, a, a2, anonymous, a01, ano01]


def test_names(session):
    assert session.names == ['a', 'a01', 'a2', 'ano01', 'anonymous', 'b', 'b024',
                             'c', 'd', 'e', 'f', 'g', 'h']
    # add them in the "wrong" order
    session.add(j='j')
    session.add(i='i')
    assert session.names == ['a', 'a01', 'a2', 'ano01', 'anonymous', 'b', 'b024',
                             'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']


def _test_io(tmpdir, session, meta, engine, ext):
    filename = f"test_{engine}.{ext}" if 'csv' not in engine else f"test_{engine}{ext}"
    fpath = tmp_path(tmpdir, filename)

    is_excel_or_csv = 'excel' in engine or 'csv' in engine

    kind = Array if is_excel_or_csv else (Axis, Group, Array) + _supported_scalars_types
    session = session.filter(kind=kind)

    session.meta = meta

    # save and load
    session.save(fpath, engine=engine)
    s = Session()
    s.load(fpath, engine=engine)
    # use Session.names instead of Session.keys because CSV, Excel and HDF do *not* keep ordering
    assert s.names == session.names
    assert s.equals(session)
    if not is_excel_or_csv:
        for key in s.filter(kind=Axis).keys():
            assert s[key].dtype == session[key].dtype
    if engine != 'pandas_excel':
        assert s.meta == meta

    # update a Group + an Axis + an array (overwrite=False)
    a4 = Axis('a=0..3')
    a4_01 = a3['0,1'] >> 'a01'
    e2 = ndtest((a4, 'b=b0..b2'))
    h2 = full_like(h, fill_value=10)
    Session(a=a4, a01=a4_01, e=e2, h=h2).save(fpath, overwrite=False, engine=engine)
    s = Session()
    s.load(fpath, engine=engine)
    if engine == 'pandas_excel':
        # Session.save() via engine='pandas_excel' always overwrite the output Excel files
        assert s.names == ['e', 'h']
    elif is_excel_or_csv:
        assert s.names == ['e', 'f', 'g', 'h']
    else:
        assert s.names == session.names
        assert s['a'].equals(a4)
        assert s['a01'].equals(a4_01)
    assert_array_nan_equal(s['e'], e2)
    if engine != 'pandas_excel':
        assert s.meta == meta

    # load only some objects
    session.save(fpath, engine=engine)
    s = Session()
    names_to_load = ['e', 'f'] if is_excel_or_csv else ['a', 'a01', 'a2', 'anonymous', 'e', 'f', 's_bool', 's_int']
    s.load(fpath, names=names_to_load, engine=engine)
    assert s.names == names_to_load
    if engine != 'pandas_excel':
        assert s.meta == meta

    return fpath


def _add_scalars_to_session(s):
    # 's' for scalar
    s['s_int'] = 5
    s['s_float'] = 5.5
    s['s_bool'] = True
    s['s_str'] = 'string'
    s['s_date'] = date(2020, 1, 10)
    s['s_time'] = time(11, 23, 54)
    s['s_datetime'] = datetime(2020, 1, 10, 11, 23, 54)
    return s


@needs_pytables
def test_h5_io(tmpdir, session, meta):
    session = _add_scalars_to_session(session)

    msg = "\nyour performance may suffer as PyTables will pickle object types"
    regex = re.compile(msg, flags=re.MULTILINE)

    # for some reason the PerformanceWarning is not detected as such, so this does not work:
    # with pytest.warns(tables.PerformanceWarning):
    with pytest.warns(Warning, match=regex):
        _test_io(tmpdir, session, meta, engine='pandas_hdf', ext='h5')


@needs_openpyxl
def test_xlsx_pandas_io(tmpdir, session, meta):
    _test_io(tmpdir, session, meta, engine='pandas_excel', ext='xlsx')


@needs_xlwings
def test_xlsx_xlwings_io(tmpdir, session, meta):
    _test_io(tmpdir, session, meta, engine='xlwings_excel', ext='xlsx')


def test_csv_io(tmpdir, session, meta):
    try:
        fpath = _test_io(tmpdir, session, meta, engine='pandas_csv', ext='csv')

        names = Session({k: v for k, v in session.items() if isinstance(v, Array)}).names

        # test loading with a pattern
        pattern = os.path.join(fpath, '*.csv')
        s = Session(pattern)
        assert s.names == names
        assert s.meta == meta

        # create an invalid .csv file
        invalid_fpath = os.path.join(fpath, 'invalid.csv')
        with open(invalid_fpath, 'w') as f:
            f.write(',",')

        # try loading the directory with the invalid file
        with pytest.raises(pd.errors.ParserError):
            s = Session(pattern)

        # test loading a pattern, ignoring invalid/unsupported files
        s = Session()
        s.load(pattern, ignore_exceptions=True)
        assert s.names == names
        assert s.meta == meta
    finally:
        shutil.rmtree(fpath)


def test_pickle_io(tmpdir, session, meta):
    session = _add_scalars_to_session(session)
    _test_io(tmpdir, session, meta, engine='pickle', ext='pkl')


def test_pickle_roundtrip(session, meta):
    original = session.filter(kind=Array)
    original.meta = meta
    s = pickle.dumps(original)
    res = pickle.loads(s)
    assert res.equals(original)
    assert res.meta == meta


def test_to_globals(session):
    msg = "Session.to_globals should usually only be used in interactive consoles and not in scripts. " \
          "Use warn=False to deactivate this warning."
    with must_warn(RuntimeWarning, msg=msg):
        session.to_globals()

    assert a is session.a
    assert b is session.b
    assert c is session.c
    assert d is session.d
    assert e is session.e
    assert f is session.f
    assert g is session.g

    # test inplace
    backup_dest = e
    backup_value = session.e.copy()
    session.e = zeros_like(e)
    session.to_globals(inplace=True, warn=False)
    # check the variable is correct (the same as before)
    assert e is backup_dest
    assert e is not session.e
    # check the content has changed
    assert_array_nan_equal(e, session.e)
    assert not e.equals(backup_value)
    # reset e to its original value
    e[:] = backup_value


def test_element_equals(session):
    session_cls = session.__class__
    other_session = session_cls([(key, value) for key, value in session.items()])

    keys = [key for key, value in session.items() if isinstance(value, (Axis, Group, Array))]
    expected_res = full(Axis(keys, 'name'), fill_value=True, dtype=bool)

    # ====== same sessions ======
    res = session.element_equals(other_session)
    assert res.axes == expected_res.axes
    assert res.equals(expected_res)

    # ====== session with missing/extra items ======
    # delete some items
    for deleted_key in ['b', 'b024', 'g']:
        del other_session[deleted_key]
        expected_res[deleted_key] = False
    # add one item
    other_session['k'] = k
    expected_res = expected_res.append('name', False, label='k')

    res = session.element_equals(other_session)
    assert res.axes == expected_res.axes
    assert res.equals(expected_res)

    # ====== session with a modified array ======
    h2 = h.copy()
    h2['a1', 'b1'] = 42
    other_session['h'] = h2
    expected_res['h'] = False

    res = session.element_equals(other_session)
    assert res.axes == expected_res.axes
    assert res.equals(expected_res)


def to_boolean_array_eq(res):
    return stack([(key, item.all() if isinstance(item, Array) else item)
                  for key, item in res.items()], 'name')


def test_eq(session):
    session_cls = session.__class__
    other_session = session_cls([(key, value) for key, value in session.items()])
    expected_res = full(Axis(list(session.keys()), 'name'), fill_value=True, dtype=bool)

    # ====== same sessions ======
    res = session == other_session
    res = to_boolean_array_eq(res)
    assert res.axes == expected_res.axes
    assert res.equals(expected_res)

    # ====== session with missing/extra items ======
    del other_session['g']
    expected_res['g'] = False
    other_session['k'] = k
    expected_res = expected_res.append('name', False, label='k')

    res = session == other_session
    res = to_boolean_array_eq(res)
    assert res.axes == expected_res.axes
    assert res.equals(expected_res)

    # ====== session with a modified array ======
    h2 = h.copy()
    h2['a1', 'b1'] = 42
    other_session['h'] = h2
    expected_res['h'] = False

    res = session == other_session
    assert res['h'].equals(session['h'] == other_session['h'])
    res = to_boolean_array_eq(res)
    assert res.axes == expected_res.axes
    assert res.equals(expected_res)


def to_boolean_array_ne(res):
    return stack([(key, item.any() if isinstance(item, Array) else item)
                  for key, item in res.items()], 'name')


def test_ne(session):
    session_cls = session.__class__
    other_session = session_cls([(key, value) for key, value in session.items()])
    expected_res = full(Axis(list(session.keys()), 'name'), fill_value=False, dtype=bool)

    # ====== same sessions ======
    res = session != other_session
    res = to_boolean_array_ne(res)
    assert res.axes == expected_res.axes
    assert res.equals(expected_res)

    # ====== session with missing/extra items ======
    del other_session['g']
    expected_res['g'] = True
    other_session['k'] = k
    expected_res = expected_res.append('name', True, label='k')

    res = session != other_session
    res = to_boolean_array_ne(res)
    assert res.axes == expected_res.axes
    assert res.equals(expected_res)

    # ====== session with a modified array ======
    h2 = h.copy()
    h2['a1', 'b1'] = 42
    other_session['h'] = h2
    expected_res['h'] = True

    res = session != other_session
    assert res['h'].equals(session['h'] != other_session['h'])
    res = to_boolean_array_ne(res)
    assert res.axes == expected_res.axes
    assert res.equals(expected_res)


def test_sub(session):
    sess = session

    # session - session
    other = Session({'e': e, 'f': f})
    other['e'] = e - 1
    other['f'] = ones_like(f)
    diff = sess - other
    assert_array_nan_equal(diff['e'], np.full((2, 3), 1, dtype=np.int32))
    assert_array_nan_equal(diff['f'], f - ones_like(f))
    assert isnan(diff['g']).all()
    assert diff.a is a
    assert diff.a01 is a01
    assert diff.c is c

    # session - scalar
    diff = sess - 2
    assert_array_nan_equal(diff['e'], e - 2)
    assert_array_nan_equal(diff['f'], f - 2)
    assert_array_nan_equal(diff['g'], g - 2)
    assert diff.a is a
    assert diff.a01 is a01
    assert diff.c is c

    # session - dict(Array and scalar)
    other = {'e': ones_like(e), 'f': 1}
    diff = sess - other
    assert_array_nan_equal(diff['e'], e - ones_like(e))
    assert_array_nan_equal(diff['f'], f - 1)
    assert isnan(diff['g']).all()
    assert diff.a is a
    assert diff.a01 is a01
    assert diff.c is c

    # session - array
    axes = [a, b]
    other = Session([('a', a), ('a01', a01), ('c', c), ('e', ndtest((a, b))),
                     ('f', full((a, b), fill_value=3)), ('g', ndtest('c=c0..c2'))])
    diff = other - ones(axes)
    assert_array_nan_equal(diff['e'], other['e'] - ones(axes))
    assert_array_nan_equal(diff['f'], other['f'] - ones(axes))
    assert_array_nan_equal(diff['g'], other['g'] - ones(axes))
    assert diff.a is a
    assert diff.a01 is a01
    assert diff.c is c


def test_rsub(session):
    sess = session

    # scalar - session
    diff = 2 - sess
    assert_array_nan_equal(diff['e'], 2 - e)
    assert_array_nan_equal(diff['f'], 2 - f)
    assert_array_nan_equal(diff['g'], 2 - g)
    assert diff.a is a
    assert diff.a01 is a01
    assert diff.c is c

    # dict(Array and scalar) - session
    other = {'e': ones_like(e), 'f': 1}
    diff = other - sess
    assert_array_nan_equal(diff['e'], ones_like(e) - e)
    assert_array_nan_equal(diff['f'], 1 - f)
    assert isnan(diff['g']).all()
    assert diff.a is a
    assert diff.a01 is a01
    assert diff.c is c


def test_div(session):
    sess = session
    session_cls = session.__class__

    other = session_cls({'e': e, 'f': f})
    other['e'] = e - 1
    other['f'] = f + 1

    with must_warn(RuntimeWarning, msg="divide by zero encountered during operation"):
        res = sess / other

    with np.errstate(divide='ignore', invalid='ignore'):
        flat_e = np.arange(6) / np.arange(-1, 5)
    assert_array_nan_equal(res['e'], flat_e.reshape(2, 3))

    flat_f = np.arange(6) / np.arange(1, 7)
    assert_array_nan_equal(res['f'], flat_f.reshape(3, 2))
    assert isnan(res['g']).all()


def test_rdiv(session):
    sess = session

    # scalar / session
    with must_warn(RuntimeWarning, msg="divide by zero encountered during operation", check_num=False):
        res = 2 / sess
    with np.errstate(divide='ignore'):
        expected_e = 2 / e
        expected_f = 2 / f
        expected_g = 2 / g
    assert_array_nan_equal(res['e'], expected_e)
    assert_array_nan_equal(res['f'], expected_f)
    assert_array_nan_equal(res['g'], expected_g)
    assert res.a is a
    assert res.a01 is a01
    assert res.c is c

    # dict(Array and scalar) - session
    other = {'e': e, 'f': f}
    with must_warn(RuntimeWarning,
                   msg="invalid value (NaN) encountered during operation (this is typically caused by a 0 / 0)",
                   check_num=False):
        res = other / sess
    with np.errstate(invalid='ignore'):
        expected_e = e / e
        expected_f = f / f
    assert_array_nan_equal(res['e'], expected_e)
    assert_array_nan_equal(res['f'], expected_f)
    assert res.a is a
    assert res.a01 is a01
    assert res.c is c


def test_local_arrays():
    h = ndtest(2)
    _h = ndtest(3)

    # exclude private local arrays
    s = local_arrays()
    s_expected = Session([('h', h)])
    assert s.equals(s_expected)

    # all local arrays
    s = local_arrays(include_private=True)
    s_expected = Session([('h', h), ('_h', _h)])
    assert s.equals(s_expected)


def test_global_arrays():
    # exclude private global arrays
    s = global_arrays()
    s_expected = Session([('e', e), ('f', f), ('g', g), ('h', h), ('k', k)])
    assert s.equals(s_expected)

    # all global arrays
    s = global_arrays(include_private=True)
    s_expected = Session([('e', e), ('_e', _e), ('f', f), ('g', g), ('h', h), ('k', k)])
    assert s.equals(s_expected)


def test_arrays():
    i = ndtest(2)
    _i = ndtest(3)

    # exclude private arrays
    s = arrays()
    s_expected = Session([('e', e), ('f', f), ('g', g), ('h', h), ('i', i), ('k', k)])
    assert s.equals(s_expected)

    # all arrays
    s = arrays(include_private=True)
    s_expected = Session([('_e', _e), ('_i', _i), ('e', e), ('f', f), ('g', g), ('h', h), ('i', i), ('k', k)])
    assert s.equals(s_expected)


if __name__ == "__main__":
    pytest.main()



__version__ = '0.33'


from larray.core.axis import Axis, AxisCollection, X
from larray.core.group import Group, LGroup, LSet, IGroup, union
from larray.core.array import (Array, zeros, zeros_like, ones, ones_like, empty, empty_like, full,
                               full_like, sequence, labels_array, ndtest, asarray, identity, diag,
                               eye, all, any, sum, prod, cumsum, cumprod, min, max, mean, ptp, var,
                               std, median, percentile, stack, zip_array_values, zip_array_items)
from larray.core.session import Session, local_arrays, global_arrays, arrays
from larray.core.checked import CheckedArray, CheckedSession, CheckedParameters
from larray.core.constants import nan, inf, pi, e, euler_gamma
from larray.core.metadata import Metadata
from larray.core.ufuncs import wrap_elementwise_array_func, maximum, minimum, where
from larray.core.npufuncs import (sin, cos, tan, arcsin, arccos, arctan, hypot, arctan2, degrees,
                                  radians, unwrap, sinh, cosh, tanh, arcsinh, arccosh, arctanh,
                                  angle, real, imag, conj,
                                  round, around, round_, rint, fix, floor, ceil, trunc,
                                  exp, expm1, exp2, log, log10, log2, log1p, logaddexp, logaddexp2,
                                  i0, sinc, signbit, copysign, frexp, ldexp,
                                  convolve, clip, sqrt, absolute, fabs, sign, fmax, fmin, nan_to_num,
                                  real_if_close, interp, isnan, isinf, inverse)
from larray.core.misc import isscalar

from larray.inout.misc import from_lists, from_string
from larray.inout.pandas import from_frame, from_series
from larray.inout.csv import read_csv, read_tsv, read_eurostat
from larray.inout.excel import read_excel
from larray.inout.hdf import read_hdf
from larray.inout.sas import read_sas
from larray.inout.stata import read_stata
from larray.inout.xw_excel import open_excel, Workbook
from larray.inout.xw_reporting import ExcelReport, ReportSheet

# module is imported only to initialize handlers for .pkl and .pickle (we del _pkl at the end of this module)
import larray.inout.pickle as _pkl

from larray.util.options import get_options, set_options

from larray.viewer import view, edit, debug, compare, run_editor_on_exception

from larray.extra.ipfp import ipfp

from larray.example import get_example_filepath, load_example_data, EXAMPLE_EXCEL_TEMPLATES_DIR


__all__ = [
    # axis
    'Axis', 'AxisCollection', 'X',
    # group
    'Group', 'LGroup', 'LSet', 'IGroup', 'union',
    # array
    'Array', 'zeros', 'zeros_like', 'ones', 'ones_like', 'empty', 'empty_like', 'full',
    'full_like', 'sequence', 'labels_array', 'ndtest', 'asarray', 'identity', 'diag', 'eye',
    'all', 'any', 'sum', 'prod', 'cumsum', 'cumprod', 'min', 'max', 'mean', 'ptp', 'var', 'std',
    'median', 'percentile', 'stack', 'zip_array_values', 'zip_array_items',
    # session
    'Session', 'local_arrays', 'global_arrays', 'arrays',
    # constrained
    'CheckedArray', 'CheckedSession', 'CheckedParameters',
    # constants
    'nan', 'inf', 'pi', 'e', 'euler_gamma',
    # metadata
    'Metadata',
    # ufuncs
    'wrap_elementwise_array_func',
    'maximum', 'minimum', 'where',
    'sin', 'cos', 'tan', 'arcsin', 'arccos', 'arctan', 'hypot', 'arctan2', 'degrees', 'radians',
    'unwrap', 'sinh', 'cosh', 'tanh', 'arcsinh', 'arccosh', 'arctanh',
    'angle', 'real', 'imag', 'conj',
    'round', 'around', 'round_', 'rint', 'fix', 'floor', 'ceil', 'trunc',
    'exp', 'expm1', 'exp2', 'log', 'log10', 'log2', 'log1p', 'logaddexp', 'logaddexp2',
    'i0', 'sinc', 'signbit', 'copysign', 'frexp', 'ldexp',
    'convolve', 'clip', 'sqrt', 'absolute', 'fabs', 'sign', 'fmax', 'fmin', 'nan_to_num',
    'real_if_close', 'interp', 'isnan', 'isinf', 'inverse',
    # core/misc
    'isscalar',
    # inout
    'from_lists', 'from_string', 'from_frame', 'from_series', 'read_csv', 'read_tsv',
    'read_eurostat', 'read_excel', 'read_hdf', 'read_sas', 'read_stata',
    'open_excel', 'Workbook', 'ExcelReport', 'ReportSheet',
    # utils
    'get_options', 'set_options',
    # viewer
    'view', 'edit', 'debug', 'compare', 'run_editor_on_exception',
    # ipfp
    'ipfp',
    # example
    'get_example_filepath', 'load_example_data', 'EXAMPLE_EXCEL_TEMPLATES_DIR',
]


# ==== DEPRECATED API ====

# F401: imported item not used
from larray.core.axis import x                                                              # noqa: F401
from larray.core.group import PGroup                                                        # noqa: F401
from larray.core.array import (LArray, aslarray, create_sequential, ndrange, larray_equal,  # noqa: F401
                               larray_nan_equal, nan_equal, element_equal)

_deprecated_names = [
    # axis
    'x',
    # group
    'PGroup',
    # array
    'LArray', 'aslarray',
    'create_sequential', 'ndrange',
    'larray_equal', 'larray_nan_equal', 'nan_equal', 'element_equal',
]

__all__ += _deprecated_names

# cleanup namespace (module was imported only to initialize handlers for .pkl and .pickle)
del _pkl



from typing import Any

import numpy as np


def isscalar(element: Any) -> bool:
    r"""
    Returns `True` if the type of element is a scalar type.

    Parameters
    ----------
    element: any
        Input argument, can be of any type and shape.

    Returns
    -------
    bool
        `True` if `element` is a scalar type, `False` if it is not.

    Examples
    --------
    >>> from larray import ndtest
    >>> isscalar(3.1)
    True
    >>> isscalar([3.1])
    False
    >>> isscalar(False)
    True
    >>> isscalar('larray')
    True

    >>> arr = ndtest((2, 2))
    >>> arr
    a\b  b0  b1
     a0   0   1
     a1   2   3
    >>> isscalar(arr)
    False
    """
    return np.isscalar(element)



from larray.util.misc import _positive_integer

DISPLAY_PRECISION = 'display_precision'
DISPLAY_WIDTH = 'display_width'
DISPLAY_MAXLINES = 'display_maxlines'
DISPLAY_EDGEITEMS = 'display_edgeitems'


_OPTIONS = {
    DISPLAY_PRECISION: None,
    DISPLAY_WIDTH: 80,
    DISPLAY_MAXLINES: 200,
    DISPLAY_EDGEITEMS: 5,
}


def _integer_maxlines(value):
    if not isinstance(value, int) and value >= -1:
        raise ValueError("Expected integer")


def _positive_integer_or_none(value):
    if value is None:
        return
    else:
        _positive_integer(value)


_VALIDATORS = {
    DISPLAY_PRECISION: _positive_integer_or_none,
    DISPLAY_WIDTH: _positive_integer,
    DISPLAY_MAXLINES: _integer_maxlines,
    DISPLAY_EDGEITEMS: _positive_integer,
}


# idea taken from xarray. See xarray/core/options.py module for original implementation.
class set_options:
    r"""Set options for larray in a controlled context.

    Currently supported options:

    - ``display_precision``: number of digits of precision for floating point output.
      Print as many digits as necessary to uniquely specify the value by default (None).
    - ``display_width``: maximum display width for ``repr`` on larray objects. Defaults to 80.
    - ``display_maxlines``: Maximum number of lines to show. All lines are shown if -1.
      Defaults to 200.
    - ``display_edgeitems`` : if number of lines to display is greater than ``display_maxlines``,
      only the first and last ``display_edgeitems`` lines are displayed.
      Only active if ``display_maxlines`` is not -1. Defaults to 5.

    Examples
    --------
    >>> from larray import *
    >>> arr = ndtest((500, 100), dtype=float) + 0.123456

    You can use ``set_options`` either as a context manager:

    >>> with set_options(display_width=100, display_edgeitems=2):
    ...     print(arr)
     a\b            b0            b1            b2  ...           b97           b98           b99
      a0      0.123456      1.123456      2.123456  ...     97.123456     98.123456     99.123456
      a1    100.123456    101.123456    102.123456  ...    197.123456    198.123456    199.123456
     ...           ...           ...           ...  ...           ...           ...           ...
    a498  49800.123456  49801.123456  49802.123456  ...  49897.123456  49898.123456  49899.123456
    a499  49900.123456  49901.123456  49902.123456  ...  49997.123456  49998.123456  49999.123456

    Or to set global options:

    >>> set_options(display_maxlines=10, display_precision=2) # doctest: +SKIP
    >>> print(arr) # doctest: +SKIP
     a\b        b0        b1        b2  ...       b97       b98       b99
      a0      0.12      1.12      2.12  ...     97.12     98.12     99.12
      a1    100.12    101.12    102.12  ...    197.12    198.12    199.12
      a2    200.12    201.12    202.12  ...    297.12    298.12    299.12
      a3    300.12    301.12    302.12  ...    397.12    398.12    399.12
      a4    400.12    401.12    402.12  ...    497.12    498.12    499.12
     ...       ...       ...       ...  ...       ...       ...       ...
    a495  49500.12  49501.12  49502.12  ...  49597.12  49598.12  49599.12
    a496  49600.12  49601.12  49602.12  ...  49697.12  49698.12  49699.12
    a497  49700.12  49701.12  49702.12  ...  49797.12  49798.12  49799.12
    a498  49800.12  49801.12  49802.12  ...  49897.12  49898.12  49899.12
    a499  49900.12  49901.12  49902.12  ...  49997.12  49998.12  49999.12

    To put back the default options, you can use:

    >>> set_options(display_precision=None, display_width=80, display_maxlines=200, display_edgeitems=5)
    ... # doctest: +SKIP
    """

    def __init__(self, **kwargs):
        self.old = {}
        for k, v in kwargs.items():
            if k not in _OPTIONS:
                raise ValueError(f'Argument {k} is not in the set of valid options {set(_OPTIONS)}')
            if k in _VALIDATORS:
                _VALIDATORS[k](v)
            self.old[k] = _OPTIONS[k]
        _OPTIONS.update(kwargs)

    def __enter__(self):
        return

    def __exit__(self, type, value, traceback):
        _OPTIONS.update(self.old)


def get_options():
    r"""
    Return the current options.

    Returns
    -------
    Dictionary of current print options with keys

        - display_precision: int or None
        - display_width: int
        - display_maxlines: int
        - display_edgeitems : int

    For a full description of these options, see :py:obj:`set_options`.

    See Also
    --------
    set_options

    Examples
    --------
    >>> get_options() # doctest: +SKIP
    {'display_precision': None, 'display_width': 80, 'display_maxlines': 200, 'display_edgeitems': 5}
    """
    return _OPTIONS.copy()



import pickle
import os.path
from collections import OrderedDict

from larray.core.axis import Axis
from larray.core.group import Group
from larray.core.array import Array
from larray.core.metadata import Metadata
from larray.inout.session import register_file_handler
from larray.inout.common import FileHandler, _supported_types, _supported_typenames, _supported_scalars_types


@register_file_handler('pickle', ['pkl', 'pickle'])
class PickleHandler(FileHandler):
    def _open_for_read(self):
        with open(self.fname, 'rb') as f:
            self.data = OrderedDict(pickle.load(f))

    def _open_for_write(self):
        if not self.overwrite_file and os.path.isfile(self.fname):
            self._open_for_read()
        else:
            self.data = OrderedDict()

    def list_items(self):
        # scalar
        items = [(key, type(value).__name__) for key, value in self.data.items()
                 if isinstance(value, _supported_scalars_types)]
        # axes
        items += [(key, 'Axis') for key, value in self.data.items() if isinstance(value, Axis)]
        # groups
        items += [(key, 'Group') for key, value in self.data.items() if isinstance(value, Group)]
        # arrays
        items += [(key, 'Array') for key, value in self.data.items() if isinstance(value, Array)]
        return items

    def _read_item(self, key, typename, *args, **kwargs):
        if typename in _supported_typenames:
            return self.data[key]
        else:
            raise TypeError()

    def _dump_item(self, key, value, *args, **kwargs):
        if isinstance(value, _supported_types):
            self.data[key] = value
        else:
            raise TypeError()

    def _read_metadata(self):
        if '__metadata__' in self.data:
            return self.data['__metadata__']
        else:
            return Metadata()

    def _dump_metadata(self, metadata):
        self.data['__metadata__'] = metadata

    def save(self):
        with open(self.fname, 'wb') as f:
            pickle.dump(self.data, f)

    def close(self):
        pass



import os

from larray import ndtest, open_excel, Session, X


DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


def generate_tests_files():
    tests = {'1d': 3,
             '2d': "a=1..3; b=b0,b1",
             '2d_classic': "a=a0..a2;b=b0..b2",
             '3d': "a=1..3; b=b0,b1; c=c0..c2",
             'int_labels': "a=0..2; b=0..2; c=0..2",
             'missing_values': "a=1..3; b=b0,b1; c=c0..c2",
             'unsorted': "a=3..1; b=b1,b0; c=c2..c0",
             'position': "a=1..3; b=b0,b1; c=c0..c2"}

    wb = open_excel(os.path.join(DATA_DIR, 'test.xlsx'), overwrite_file=True)
    wb_narrow = open_excel(os.path.join(DATA_DIR, 'test_narrow.xlsx'), overwrite_file=True)

    for name, dim in tests.items():
        arr = ndtest(dim)
        if name == '2d_classic':
            df = arr.to_frame(fold_last_axis_name=False)
            # wide format
            df.to_csv(os.path.join(DATA_DIR, f'test{name}.csv'), sep=',', na_rep='')
            wb[name] = ''
            wb[name]['A1'].options().value = df
            # narrow format
            df = arr.to_series(name='value')
            df.to_csv(os.path.join(DATA_DIR, f'test{name}_narrow.csv'), sep=',', na_rep='', header=True)
            wb_narrow[name] = ''
            wb_narrow[name]['A1'].options().value = df
        elif name == 'missing_values':
            df = arr.to_frame(fold_last_axis_name=True)
            # wide format
            df = df.drop([(2, 'b0'), (3, 'b1')])
            df.to_csv(os.path.join(DATA_DIR, f'test{name}.csv'), sep=',', na_rep='')
            wb[name] = ''
            wb[name]['A1'].options().value = df
            # narrow format
            df = arr.to_series(name='value')
            df = df.drop([(2, 'b0'), (2, 'b1', 'c1'), (3, 'b1')])
            df.to_csv(os.path.join(DATA_DIR, f'test{name}_narrow.csv'), sep=',', na_rep='', header=True)
            wb_narrow[name] = ''
            wb_narrow[name]['A1'].options().value = df
        elif name == 'position':
            # wide format
            wb[name] = ''
            wb[name]['D3'] = arr.dump()
            # narrow format
            wb_narrow[name] = ''
            wb_narrow[name]['D3'] = arr.dump(wide=False)
        else:
            # wide format
            arr.to_csv(os.path.join(DATA_DIR, f'test{name}.csv'))
            wb[name] = arr.dump()
            # narrow format
            arr.to_csv(os.path.join(DATA_DIR, f'test{name}_narrow.csv'), wide=False)
            wb_narrow[name] = arr.dump(wide=False)

    wb.save()
    wb.close()
    wb_narrow.save()
    wb_narrow.close()


def generate_example_files(csv=True, excel=True, hdf5=True):
    from larray_eurostat import eurostat_get

    def prepare_eurostat_data(dataset_name, countries):
        arr = eurostat_get(dataset_name)[X.unit['NR'], X.age['TOTAL'], X.sex['M,F']]
        arr = arr[X.time[::-1]][2013:2017]
        arr = arr.rename('sex', 'gender')
        arr = arr.set_labels(gender='Male,Female')
        arr = arr.rename('geo', 'country')
        country_codes = list(countries.keys())
        country_names = list(countries.values())
        if dataset_name == 'migr_imm1ctz':
            # example of an array with ambiguous axes
            arr = arr['COMPLET', X.citizen[country_codes], X.country[country_codes]].astype(int)
            arr = arr.rename('citizen', 'citizenship')
            arr = arr.set_labels('citizenship', country_names)
            arr = arr.set_labels('country', country_names)
            arr = arr.transpose('country', 'citizenship', 'gender', 'time')
        else:
            arr = arr[country_codes].astype(int)
            arr = arr.set_labels('country', country_names)
            arr = arr.transpose('country', 'gender', 'time')
        return arr

    countries = {'BE': 'Belgium', 'FR': 'France', 'DE': 'Germany'}
    benelux = {'BE': 'Belgium', 'LU': 'Luxembourg', 'NL': 'Netherlands'}

    # Arrays
    population = prepare_eurostat_data('demo_pjan', countries)
    population.meta.title = 'Population on 1 January by age and sex'
    population.meta.source = 'table demo_pjan from Eurostat'
    # ----
    population_benelux = prepare_eurostat_data('demo_pjan', benelux)
    population_benelux.meta.title = 'Population on 1 January by age and sex (Benelux)'
    population_benelux.meta.source = 'table demo_pjan from Eurostat'
    # ----
    population_5_countries = population.extend('country', population_benelux[['Luxembourg', 'Netherlands']])
    population_5_countries.meta.title = 'Population on 1 January by age and sex (Benelux + France + Germany)'
    population_5_countries.meta.source = 'table demo_pjan from Eurostat'
    # ----
    births = prepare_eurostat_data('demo_fasec', countries)
    births.meta.title = "Live births by mother's age and newborn's sex"
    births.meta.source = 'table demo_fasec from Eurostat'
    # ----
    deaths = prepare_eurostat_data('demo_magec', countries)
    deaths.meta.title = 'Deaths by age and sex'
    deaths.meta.source = 'table demo_magec from Eurostat'
    # ----
    immigration = prepare_eurostat_data('migr_imm1ctz', benelux)
    immigration.meta.title = 'Immigration by age group, sex and citizenship'
    immigration.meta.source = 'table migr_imm1ctz from Eurostat'

    # Groups
    even_years = population.time[2014::2] >> 'even_years'
    odd_years = population.time[2013::2] >> 'odd_years'

    # Session
    ses = Session({'country': population.country, 'country_benelux': immigration.country,
                   'citizenship': immigration.citizenship,
                   'gender': population.gender, 'time': population.time,
                   'even_years': even_years, 'odd_years': odd_years,
                   'population': population, 'population_benelux': population_benelux,
                   'population_5_countries': population_5_countries,
                   'births': births, 'deaths': deaths, 'immigration': immigration})
    ses.meta.title = 'Demographic datasets for a small selection of countries in Europe'
    ses.meta.source = 'demo_jpan, demo_fasec, demo_magec and migr_imm1ctz tables from Eurostat'

    # EUROSTAT DATASET

    if csv:
        ses.save(os.path.join(DATA_DIR, 'demography_eurostat'))
    if excel:
        ses.save(os.path.join(DATA_DIR, 'demography_eurostat.xlsx'))
    if hdf5:
        ses.save(os.path.join(DATA_DIR, 'demography_eurostat.h5'))

    # EXAMPLE FILES

    years = population.time[2013:2015]
    population = population[years]
    population_narrow = population['Belgium,France'].sum('gender')
    births = births[years]
    deaths = deaths[years]
    immigration = immigration[years]

    # Dataframes (for testing missing axis/values)
    df_missing_axis_name = population.to_frame(fold_last_axis_name=False)
    df_missing_values = population.to_frame(fold_last_axis_name=True)
    df_missing_values.drop([('France', 'Male'), ('Germany', 'Female')], inplace=True)

    if csv:
        examples_dir = os.path.join(DATA_DIR, 'examples')
        population.to_csv(os.path.join(examples_dir, 'population.csv'))
        births.to_csv(os.path.join(examples_dir, 'births.csv'))
        deaths.to_csv(os.path.join(examples_dir, 'deaths.csv'))
        immigration.to_csv(os.path.join(examples_dir, 'immigration.csv'))
        df_missing_axis_name.to_csv(os.path.join(examples_dir, 'population_missing_axis_name.csv'), sep=',', na_rep='')
        df_missing_values.to_csv(os.path.join(examples_dir, 'population_missing_values.csv'), sep=',', na_rep='')
        population_narrow.to_csv(os.path.join(examples_dir, 'population_narrow_format.csv'), wide=False)

    if excel:
        with open_excel(os.path.join(DATA_DIR, 'examples.xlsx'), overwrite_file=True) as wb:
            wb['population'] = population.dump()
            wb['births'] = births.dump()
            wb['deaths'] = deaths.dump()
            wb['immigration'] = immigration.dump()
            wb['population_births_deaths'] = population.dump()
            wb['population_births_deaths']['A9'] = births.dump()
            wb['population_births_deaths']['A17'] = deaths.dump()
            wb['population_missing_axis_name'] = ''
            wb['population_missing_axis_name']['A1'].options().value = df_missing_axis_name
            wb['population_missing_values'] = ''
            wb['population_missing_values']['A1'].options().value = df_missing_values
            # wb['population_narrow_format'] = population_narrow.dump(wide=False)
            wb.save()
        population_narrow.to_excel(os.path.join(DATA_DIR, 'examples.xlsx'), 'population_narrow_format', wide=False)
        Session({'country': population.country, 'gender': population.gender, 'time': population.time,
                 'population': population}).save(os.path.join(DATA_DIR, 'population_only.xlsx'))
        Session({'births': births, 'deaths': deaths}).save(os.path.join(DATA_DIR, 'births_and_deaths.xlsx'))

    if hdf5:
        examples_h5_file = os.path.join(DATA_DIR, 'examples.h5')
        population.to_hdf(examples_h5_file, 'population')
        births.to_hdf(examples_h5_file, 'births')
        deaths.to_hdf(examples_h5_file, 'deaths')
        immigration.to_hdf(examples_h5_file, 'immigration')


if __name__ == '__main__':
    # generate_tests_files()
    generate_example_files()



# Portions (part of the docstrings) of this file come from numpy/random/mtrand/mtrand.pyx
# that file (and thus those portions) are licensed under the terms below

# mtrand.pyx -- A Pyrex wrapper of Jean-Sebastien Roy's RandomKit
#
# Copyright 2005 Robert Kern (robert.kern@gmail.com)
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import numpy as np

from larray.core.axis import Axis, AxisCollection       # noqa: F401
from larray.core.array import Array, asarray
from larray.core.array import raw_broadcastable
import larray as la                                     # noqa: F401


__all__ = ['randint', 'normal', 'uniform', 'permutation', 'choice']


def generic_random(np_func, args, min_axes, meta):
    args, res_axes = raw_broadcastable(args, min_axes=min_axes)
    res_data = np_func(*args, size=res_axes.shape)
    return Array(res_data, res_axes, meta=meta)


# We choose to place the axes argument in place of the numpy size argument, instead of having axes as the first
# argument, because that would make it ugly for scalars. As a consequence, it is slightly ugly when arguments
# before axes are not required.
def randint(low, high=None, axes=None, dtype='l', meta=None):
    r"""Return random integers from `low` (inclusive) to `high` (exclusive).

    Return random integers from the "discrete uniform" distribution of the specified dtype in the "half-open" interval
    [`low`, `high`). If `high` is None (the default), then results are from [0, `low`).

    Parameters
    ----------
    low : int
        Lowest (signed) integer to be drawn from the distribution (unless ``high=None``, in which case this parameter
        is one above the *highest* such integer).
    high : int, optional
        If provided, one above the largest (signed) integer to be drawn from the distribution (see above for behavior
        if ``high=None``).
    axes : int, tuple of int, str, Axis or tuple/list/AxisCollection of Axis, optional
        Axes (or shape) of the resulting array. If ``axes`` is None (the default), a single value is returned.
        Otherwise, if the resulting axes have a shape of, e.g., ``(m, n, k)``, then ``m * n * k`` samples are drawn.
    dtype : data-type, optional
        Desired dtype of the result. All dtypes are determined by their name, i.e., 'int64', 'int', etc, so byteorder
        is not available and a specific precision may have different C types depending on the platform.
        The default value is 'np.int'.
    meta : list of pairs or dict or OrderedDict or Metadata, optional
        Metadata (title, description, author, creation_date, ...) associated with the array.
        Keys must be strings. Values must be of type string, int, float, date, time or datetime.

    Returns
    -------
    Array

    Examples
    --------
    Generate a single int between 0 and 9, inclusive:

    >>> la.random.randint(10)                                   # doctest: +SKIP
    6

    Generate an array of 10 ints between 1 and 5, inclusive:

    >>> la.random.randint(1, 6, 10)                             # doctest: +SKIP
    {0}*  0  1  2  3  4  5  6  7  8  9
          1  1  5  1  5  4  3  4  2  1

    Generate a 2 x 3 array of ints between 0 and 4, inclusive:

    >>> la.random.randint(5, axes=(2, 3))                       # doctest: +SKIP
    {0}*\{1}*  0  1  2
            0  4  4  1
            1  1  2  2
    >>> la.random.randint(5, axes='a=a0,a1;b=b0..b2')           # doctest: +SKIP
    a\b  b0  b1  b2
     a0   0   3   1
     a1   4   0   1

    With varying low and high (each depending on a different axis)

    >>> low = la.sequence('a=a0,a1')
    >>> low
    a  a0  a1
        0   1
    >>> high = la.sequence('b=b0..b2', initial=3)
    >>> high
    b  b0  b1  b2
        3   4   5

    In other words, we want to generate values between low and high (high included) for each cell. Let's
    note that low..high:

    a\b    b0    b1    b2
     a0  0..2  0..3  0..4
     a1  1..2  1..3  1..4

    >>> la.random.randint(low, high)                            # doctest: +SKIP
    a\b  b0  b1  b2
     a0   0   2   2
     a1   2   3   4
    """
    return generic_random(np.random.randint, (low, high), axes, meta)


def normal(loc=0.0, scale=1.0, axes=None, meta=None):
    r"""
    Draw random samples from a normal (Gaussian) distribution.

    Its probability density function is often called the bell curve because of its characteristic shape (see the
    example below)

    Parameters
    ----------
    loc : float or array_like of floats
        Mean ("centre") of the distribution.
    scale : float or array_like of floats
        Standard deviation (spread or "width") of the distribution.
    axes : int, tuple of int, str, Axis or tuple/list/AxisCollection of Axis, optional
        Minimum axes the resulting array must have. Defaults to None. The resulting array axes will be the union of
        those mentioned in ``axes`` and those of ``loc`` and ``scale``. If ``loc`` and ``scale`` are scalars and
        ``axes`` is None, a single value is returned. Otherwise, if the resulting axes have a shape of, e.g.,
        ``(m, n, k)``, then ``m * n * k`` samples are drawn.
    meta : list of pairs or dict or OrderedDict or Metadata, optional
        Metadata (title, description, author, creation_date, ...) associated with the array.
        Keys must be strings. Values must be of type string, int, float, date, time or datetime.

    Returns
    -------
    Array or scalar
        Drawn samples from the parameterized normal distribution.

    Notes
    -----
    The normal distributions occurs often in nature.  For example, it describes the commonly occurring distribution of
    samples influenced by a large number of tiny, random disturbances, each with its own unique distribution [2]_.

    The probability density function for the Gaussian distribution, first derived by De Moivre and 200 years later by
    both Gauss and Laplace independently [2]_, is

    .. math:: p(x) = \frac{1}{\sqrt{ 2 \pi \sigma^2 }}
                     e^{ - \frac{ (x - \mu)^2 } {2 \sigma^2} },

    where :math:`\mu` is the mean and :math:`\sigma` the standard deviation. The square of the standard deviation,
    :math:`\sigma^2`, is called the variance.

    The function has its peak at the mean, and its "spread" increases with the standard deviation (the function reaches
    0.607 times its maximum at :math:`x + \sigma` and :math:`x - \sigma` [2]_).  This implies that
    `la.random.normal` is more likely to return samples lying close to the mean, rather than those far away.

    References
    ----------
    .. [1] Wikipedia, "Normal distribution",
           http://en.wikipedia.org/wiki/Normal_distribution
    .. [2] P. R. Peebles Jr., "Central Limit Theorem" in "Probability,
           Random Variables and Random Signal Principles", 4th ed., 2001, pp. 51, 51, 125.

    Examples
    --------
    Generate a 2 x 3 array with numbers drawn from the distribution:

    >>> la.random.normal(0, 1, axes=(2, 3))                                         # doctest: +SKIP
    {0}*\{1}*                   0                     1                   2
            0  0.3564325741877542    0.8944149721039006  1.7206904920773107
            1  0.6904447654719367  -0.09395966570976753   0.185136309092257

    With named and labelled axes

    >>> la.random.normal(0, 1, axes='a=a0,a1;b=b0..b2')                             # doctest: +SKIP
    a\b                  b0                   b1                   b2
     a0  2.3096106652701827  -0.4269082412118316  -1.0862791566867225
     a1  0.8598817639620348   -2.386411240813283  0.10116503197279443

    With varying loc and scale (each depending on a different axis)

    >>> mu = la.sequence('a=a0,a1', initial=5, inc=5)
    >>> mu
    a  a0  a1
        5  10
    >>> sigma = la.sequence('b=b0..b2', initial=1)
    >>> sigma
    b  b0  b1  b2
        1   2   3
    >>> la.random.normal(mu, sigma)                                                 # doctest: +SKIP
    a\b                  b0                  b1                  b2
     a0   5.939369790854615  2.5043856460438403    8.33560126941519
     a1  10.759526714752091  10.093213549397403  11.705881778249683

    Draw 1000 samples from the distribution:

    >>> mu, sigma = 0, 0.1  # mean and standard deviation
    >>> sample = la.random.normal(mu, sigma, 1000)

    Verify the mean and the variance:

    >>> abs(mu - la.mean(sample)) < 0.01
    True
    >>> abs(sigma - la.std(sample, ddof=1)) < 0.01
    True

    Display the histogram of the samples, along with the probability density function:

    >>> import matplotlib.pyplot as plt                                         # doctest: +SKIP
    >>> count, bins, ignored = plt.hist(sample, 30, normed=True)                # doctest: +SKIP
    >>> pdf = 1 / (sigma * la.sqrt(2 * la.pi)) \
    ...       * la.exp(- (bins - mu) ** 2 / (2 * sigma ** 2))                   # doctest: +SKIP
    >>> _ = plt.plot(bins, pdf, linewidth=2, color='r')                         # doctest: +SKIP
    >>> plt.show()                                                              # doctest: +SKIP
    """
    return generic_random(np.random.normal, (loc, scale), axes, meta)


def uniform(low=0.0, high=1.0, axes=None, meta=None):
    r"""
    Draw samples from a uniform distribution.

    Samples are uniformly distributed over the half-open interval ``[low, high)`` (includes low, but excludes high).
    In other words, any value within the given interval is equally likely to be drawn by `uniform`.

    Parameters
    ----------
    low : float or array_like of floats, optional
        Lower boundary of the output interval.  All values generated will be greater than or equal to low.
        Defaults to 0.0.
    high : float or array_like of floats, optional
        Upper boundary of the output interval.  All values generated will be less than high.
        Defaults to 1.0.
    axes : int, tuple of int, str, Axis or tuple/list/AxisCollection of Axis, optional
        Minimum axes the resulting array must have. Defaults to None. The resulting array axes will be the union of
        those mentioned in ``axes`` and those of ``low`` and ``high``. If ``low`` and ``high`` are scalars and
        ``axes`` is None, a single value is returned. Otherwise, if the resulting axes have a shape of, e.g.,
        ``(m, n, k)``, then ``m * n * k`` samples are drawn.
    meta : list of pairs or dict or OrderedDict or Metadata, optional
        Metadata (title, description, author, creation_date, ...) associated with the array.
        Keys must be strings. Values must be of type string, int, float, date, time or datetime.

    Returns
    -------
    Array or scalar
        Drawn samples from the parameterized uniform distribution.

    See Also
    --------
    randint : Discrete uniform distribution, yielding integers.

    Notes
    -----
    The probability density function of the uniform distribution is

    .. math:: p(x) = \frac{1}{b - a}

    anywhere within the interval ``[a, b)``, and zero elsewhere.

    When ``high`` == ``low``, values of ``low`` will be returned.
    If ``high`` < ``low``, the results are officially undefined and may eventually raise an error, i.e. do not rely on
    this function to behave when passed arguments satisfying that inequality condition.

    Examples
    --------
    Generate a single sample from the distribution:

    >>> la.random.uniform()                                                          # doctest: +SKIP
    0.4616049008844396

    Generate a 2 x 3 array with numbers drawn from the distribution:

    >>> la.random.uniform(0, 5, axes=(2, 3))                                         # doctest: +SKIP
    {0}*\{1}*                   0                  1                  2
            0  3.4951791043804192  3.888533056628081  4.347461073315136
            1   2.146211610940853  0.509146487437932  2.790852715735223

    With named and labelled axes

    >>> la.random.uniform(1, 2, axes='a=a0,a1;b=b0..b2')                             # doctest: +SKIP
    a\b                  b0                  b1                  b2
     a0  1.4167729850467825  1.6953091052066793  1.2321770607672526
     a1  1.4386221912579358  1.8480607144284926  1.1726213637670433

    With varying low and high (each depending on a different axis)

    >>> low = la.sequence('a=a0,a1')
    >>> low
    a  a0  a1
        0   1
    >>> high = la.sequence('b=b0..b2', initial=1, inc=0.5)
    >>> high
    b   b0   b1   b2
       1.0  1.5  2.0
    >>> la.random.uniform(low, high)                                                 # doctest: +SKIP
    a\b                   b0                  b1                  b2
     a0  0.44608671494167573   0.948315996350121    1.74189664009661
     a1                  1.0  1.1099944474264194  1.1362792569316835

    Draw 1000 samples from the distribution:

    >>> s = la.random.uniform(-1, 0, 1000)

    All values are within the given interval:

    >>> la.all(s >= -1)
    True
    >>> la.all(s < 0)
    True

    Display the histogram of the samples, along with the probability density function:

    >>> import matplotlib.pyplot as plt                                     # doctest: +SKIP
    >>> count, bins, ignored = plt.hist(s, 15, normed=True)                 # doctest: +SKIP
    >>> _ = plt.plot(bins, np.ones_like(bins), linewidth=2, color='r')      # doctest: +SKIP
    >>> plt.show()                                                          # doctest: +SKIP
    """
    return generic_random(np.random.uniform, (low, high), axes, meta)


def permutation(x, axis=0):
    r"""
    Randomly permute a sequence along an axis, or return a permuted range.

    Parameters
    ----------
    x : int or array_like
        If `x` is an integer, randomly permute ``sequence(x)``.
        If `x` is an array, returns a randomly shuffled copy.
    axis : int, str or Axis, optional
        Axis along which to permute. Defaults to the first axis.

    Returns
    -------
    Array
        Permuted sequence or array range.

    Examples
    --------
    >>> la.random.permutation(10)                               # doctest: +SKIP
    {0}*  0  1  2  3  4  5  6  7  8  9
          6  8  0  9  4  7  1  5  3  2
    >>> la.random.permutation([1, 4, 9, 12, 15])                # doctest: +SKIP
    {0}*  0   1   2  3  4
          1  15  12  9  4
    >>> la.random.permutation(la.ndtest(5))                     # doctest: +SKIP
    a  a3  a1  a2  a4  a0
        3   1   2   4   0
    >>> arr = la.ndtest((3, 3))                                 # doctest: +SKIP
    >>> la.random.permutation(arr)                              # doctest: +SKIP
    a\b  b0  b1  b2
     a1   3   4   5
     a2   6   7   8
     a0   0   1   2
    >>> la.random.permutation(arr, axis='b')                    # doctest: +SKIP
    a\b  b1  b2  b0
     a0   1   2   0
     a1   4   5   3
     a2   7   8   6
    """
    if isinstance(x, (int, np.integer)):
        return Array(np.random.permutation(x))
    else:
        x = asarray(x)
        axis = x.axes[axis]
        g = axis.i[np.random.permutation(len(axis))]
        return x[g]


def choice(choices=None, axes=None, replace=True, p=None, meta=None):
    r"""
    Generates a random sample from given choices

    Parameters
    ----------
    choices : 1-D array-like or int, optional
        Values to choose from.
        If an array, a random sample is generated from its elements.
        If an int n, the random sample is generated as if choices was la.sequence(n)
        If p is a 1-D Array, choices are taken from its axis.
    axes : int, tuple of int, str, Axis or tuple/list/AxisCollection of Axis, optional
        Axes (or shape) of the resulting array. If ``axes`` is None (the default), a single value is returned.
        Otherwise, if the resulting axes have a shape of, e.g., ``(m, n, k)``, then ``m * n * k`` samples are drawn.
    replace : boolean, optional
        Whether the sample is with or without replacement.
    p : array-like, optional
        The probabilities associated with each entry in choices.
        If p is a 1-D Array, choices are taken from its axis labels. If p is an N-D Array, each cell represents the
        probability that the combination of labels will occur.
        If not given the sample assumes a uniform distribution over all entries in choices.
    meta : list of pairs or dict or OrderedDict or Metadata, optional
        Metadata (title, description, author, creation_date, ...) associated with the array.
        Keys must be strings. Values must be of type string, int, float, date, time or datetime.

    Returns
    -------
    Array or scalar
        The generated random samples with given ``axes`` (or shape).

    Raises
    ------
    ValueError
        If choices is an int and less than zero, if choices or p are not 1-dimensional,
        if choices is an array-like of size 0, if p is not a vector of probabilities,
        if choices and p have different lengths, or
        if replace=False and the sample size is greater than the population size.

    See Also
    --------
    randint, permutation

    Examples
    --------
    Generate one random value out of given choices (each choice has the same probability of occurring):

    >>> la.random.choice(['hello', 'world', '!'])                                       # doctest: +SKIP
    hello

    With given probabilities:

    >>> la.random.choice(['hello', 'world', '!'], p=[0.1, 0.8, 0.1])                    # doctest: +SKIP
    world

    Generate a 2 x 3 array with given axes and values drawn from the given choices using given probabilities:

    >>> la.random.choice([5, 10, 15], p=[0.3, 0.5, 0.2], axes='a=a0,a1;b=b0..b2')       # doctest: +SKIP
    a\b  b0  b1  b2
     a0  15  10  10
     a1  10   5  10

    Same as above with labels and probabilities given as a one dimensional Array

    >>> proba = Array([0.3, 0.5, 0.2], Axis([5, 10, 15], 'outcome'))                   # doctest: +SKIP
    >>> proba                                                                           # doctest: +SKIP
    outcome    5   10   15
             0.3  0.5  0.2
    >>> choice(p=proba, axes='a=a0,a1;b=b0..b2')                                        # doctest: +SKIP
    a\b  b0  b1  b2
     a0  10  15   5
     a1  10   5  10

    Generate a uniform random sample of size 3 from la.sequence(5):

    >>> la.random.choice(5, 3)                                                          # doctest: +SKIP
    {0}*  0  1  2
          3  2  0
    >>> # This is equivalent to la.random.randint(0, 5, 3)

    Generate a non-uniform random sample of size 3 from the given choices without replacement:

    >>> la.random.choice(['hello', 'world', '!'], 3, replace=False, p=[0.1, 0.6, 0.3])  # doctest: +SKIP
    {0}*      0  1      2
          world  !  hello

    Using an N-dimensional array as probabilities:

    >>> proba = Array([[0.15, 0.25, 0.10],
    ...                 [0.20, 0.10, 0.20]], 'a=a0,a1;b=b0..b2')                        # doctest: +SKIP
    >>> proba                                                                           # doctest: +SKIP
    a\b    b0    b1   b2
     a0  0.15  0.25  0.1
     a1   0.2   0.1  0.2
    >>> choice(p=proba, axes='draw=d0..d5')                                             # doctest: +SKIP
    draw\axis   a   b
           d0  a1  b2
           d1  a1  b1
           d2  a0  b1
           d3  a0  b0
           d4  a1  b2
           d5  a0  b1
    """
    axes = AxisCollection(axes)
    if isinstance(p, Array):
        if choices is not None:
            raise ValueError("choices argument cannot be used when p argument is an Array")

        if p.ndim > 1:
            flat_p = p.data.reshape(-1)
            flat_indices = choice(p.size, axes=axes, replace=replace, p=flat_p)
            return p.axes._flat_lookup(flat_indices)
        else:
            choices = p.axes[0].labels
            p = p.data
    if choices is None:
        raise ValueError("choices argument must be provided unless p is an Array")
    return Array(np.random.choice(choices, axes.shape, replace, p), axes, meta=meta)




handler_classes = {}
ext_default_engine = {}


def register_file_handler(engine, extensions=None):
    r"""Class decorator to register new file handler class

    Parameters
    ----------
    engine : str
        Engine name associated with the file handler.
    extensions : str or list of str, optional
        Extension(s) associated with the file handler.
    """
    def decorate_class(cls):
        if engine not in handler_classes:
            handler_classes[engine] = cls
        if extensions is None:
            exts = []
        elif isinstance(extensions, str):
            exts = [extensions]
        else:
            exts = extensions
        for ext in exts:
            ext_default_engine[ext] = engine
        return cls
    return decorate_class


def get_file_handler(engine):
    if engine not in handler_classes:
        raise TypeError(f"Engine {engine} is currently not implemented")
    file_handler_cls = handler_classes[engine]
    return file_handler_cls



import re
import os
import faulthandler
from pathlib import Path

import pytest
import numpy as np

from larray.tests.common import needs_xlwings, needs_pytables, must_warn
from larray import ndtest, open_excel, asarray, Axis, nan, ExcelReport, read_excel
from larray.inout import xw_excel
from larray.example import load_example_data, EXAMPLE_EXCEL_TEMPLATES_DIR


TEST_DATA_PATH = Path(__file__).parent / 'data'


@needs_xlwings
class TestWorkbook:
    def test_open_excel(self):
        import pywintypes

        # not using context manager because we call .quit manually
        wb1 = open_excel(visible=False)
        app1 = wb1.app
        # close workbook but leave app instance open (anything using wb1 will fail now)
        wb1.close()

        # disable faulthandler to avoid annoying "Windows fatal exception" messages in the console
        # See https://github.com/pywinauto/pywinauto/issues/858
        # and https://stackoverflow.com/questions/57523762/pytest-windows-fatal-exception-code-0x8001010d
        faulthandler_enabled = faulthandler.is_enabled()
        if faulthandler_enabled:
            faulthandler.disable()
        with pytest.raises(pywintypes.com_error):
            wb1.sheet_names()
        if faulthandler_enabled:
            faulthandler.enable()

        wb2 = open_excel(visible=False)
        app2 = wb2.app
        assert app1 == app2 == xw_excel.global_app
        # this effectively close all workbooks but leaves the instance intact (this is probably due to us keeping a
        # reference to it).
        app1.quit()
        # anything using wb2 will fail
        if faulthandler_enabled:
            faulthandler.disable()
        with pytest.raises(pywintypes.com_error):
            wb2.sheet_names()
        if faulthandler_enabled:
            faulthandler.enable()

        # in any case, this should work
        with open_excel(visible=False) as wb:
            wb['test'] = 'content'

    def test_links(self):
        data_dir = TEST_DATA_PATH / 'excel_with_links'
        fpath1 = data_dir / 'BookA.xlsx'
        fpath2 = data_dir / 'BookB.xlsx'
        fpath3 = data_dir / 'BookC.xlsx'
        assert read_excel(fpath1)['link to cell with link to other workbook', 'value from link'] == 4
        assert read_excel(fpath2)['cell with link to other workbook', 'formula value'] == 4
        assert read_excel(fpath3).i[0, 0] == 3

        with open_excel(fpath1) as a, open_excel(fpath2) as b, open_excel(fpath3) as c:
            c[0]['B2'] = 41
            a.save()
            b.save()
            c.save()
        assert read_excel(fpath1)['link to cell with link to other workbook', 'value from link'] == 42
        assert read_excel(fpath2)['cell with link to other workbook', 'formula value'] == 42
        assert read_excel(fpath3).i[0, 0] == 41
        with open_excel(fpath1) as a, open_excel(fpath2) as b, open_excel(fpath3) as c:
            c[0]['B2'] = 3
            a.save()
            b.save()
            c.save()
        assert read_excel(fpath1)['link to cell with link to other workbook', 'value from link'] == 4
        assert read_excel(fpath2)['cell with link to other workbook', 'formula value'] == 4
        assert read_excel(fpath3).i[0, 0] == 3

    def test_repr(self):
        with open_excel(visible=False) as wb:
            assert re.match(r'<larray.inout.xw_excel.Workbook \[Book\d+]>', repr(wb))

    def test_getitem(self):
        with open_excel(visible=False) as wb:
            sheet = wb[0]
            assert isinstance(sheet, xw_excel.Sheet)
            # this might not be true on non-English locale
            assert sheet.name == 'Sheet1'

            # this might not work on non-English locale
            sheet = wb['Sheet1']
            assert isinstance(sheet, xw_excel.Sheet)
            assert sheet.name == 'Sheet1'

            with pytest.raises(KeyError, match="Workbook has no sheet named this_sheet_does_not_exist"):
                _ = wb['this_sheet_does_not_exist']

    def test_setitem(self):
        with open_excel(visible=False) as wb:
            # sheet did not exist, str value
            wb['sheet1'] = 'sheet1 content'
            wb['sheet2'] = 'sheet2 content'
            assert wb.sheet_names() == ['sheet1', 'sheet2']

            # sheet did exist, str value
            wb['sheet2'] = 'sheet2 content v2'
            assert wb.sheet_names() == ['sheet1', 'sheet2']
            assert wb['sheet2']['A1'].value == 'sheet2 content v2'

            # sheet did not exist, Sheet value
            wb['sheet3'] = wb['sheet1']
            assert wb.sheet_names() == ['sheet1', 'sheet2', 'sheet3']
            assert wb['sheet3']['A1'].value == 'sheet1 content'

            # sheet did exist, Sheet value
            wb['sheet2'] = wb['sheet1']
            assert wb.sheet_names() == ['sheet1', 'sheet2', 'sheet3']
            assert wb['sheet2']['A1'].value == 'sheet1 content'

            with open_excel(visible=False, app="new") as wb2:
                assert wb.app != wb2.app
                with pytest.raises(ValueError, match="cannot copy a sheet from one instance of Excel to another"):
                    wb2['sheet1'] = wb['sheet1']

            # group key
            arr = ndtest((3, 3))
            for label in arr.b:
                wb[label] = arr[label].dump()
                assert arr[label].equals(wb[label].load())

    def test_delitem(self):
        with open_excel(visible=False) as wb:
            wb['sheet1'] = 'sheet1 content'
            wb['sheet2'] = 'sheet2 content'
            del wb['sheet1']
            assert wb.sheet_names() == ['sheet2']

    def test_rename(self):
        with open_excel(visible=False) as wb:
            wb['sheet_name'] = 'sheet content'
            wb['sheet_name'].name = 'renamed'
            assert wb.sheet_names() == ['renamed']


@needs_xlwings
class TestSheet:
    def test_get_and_set_item(self):
        arr = ndtest((2, 3))

        with open_excel(visible=False) as wb:
            sheet = wb[0]
            # set a few values
            sheet['A1'] = 1.5
            sheet['A2'] = 2
            sheet['A3'] = True
            sheet['A4'] = 'toto'
            # array without header
            sheet['A5'] = arr
            # array with header
            sheet['A8'] = arr.dump()

            # object array with a *numpy* NaN (dc2019 infamous 65535 bug)
            obj_arr = ndtest((2, 3)).astype(object)
            obj_arr['a0', 'b1'] = np.float64('nan')

            assert type(obj_arr['a0', 'b1']) is np.float64

            obj_arr_dump = obj_arr.dump()
            # [['a\\b', 'b0', 'b1', 'b2'], ['a0', 0, nan, 2], ['a1', 3, 4, 5]]

            # float and *not* np.float64, otherwise it gets converted to 65535 when written to Excel
            assert type(obj_arr_dump[1][2]) is float

            sheet['A12'] = obj_arr_dump

            # read them back
            assert sheet['A1'].value == 1.5
            assert sheet['A2'].value == 2
            assert sheet['A3'].value is True
            assert sheet['A4'].value == 'toto'
            # array without header
            assert np.array_equal(sheet['A5:C6'].value, arr.data)
            # array with header
            assert sheet['A8:D10'].load().equals(arr)
            assert sheet['A12:D14'].load().equals(obj_arr, nans_equal=True)

    def test_asarray(self):
        with open_excel(visible=False) as wb:
            sheet = wb[0]

            arr1 = ndtest((2, 3))
            # no header so that we have an uniform dtype for the whole sheet
            sheet['A1'] = arr1
            res1 = np.asarray(sheet)
            assert np.array_equal(res1, arr1.data)
            assert res1.dtype == arr1.dtype

    def test_array_method(self):
        with open_excel(visible=False) as wb:
            sheet = wb[0]

            # normal test array
            arr1 = ndtest((2, 3))
            sheet['A1'] = arr1.dump()
            res1 = sheet.array('B2:D3', 'A2:A3', 'B1:D1', names=['a', 'b'])
            assert arr1.equals(res1)

            # array with int labels
            arr2 = ndtest('0..1;0..2')
            sheet['A1'] = arr2.dump()
            res2 = sheet.array('B2:D3', 'A2:A3', 'B1:D1')
            # larray_equal passes even if the labels are floats...
            assert arr2.equals(res2)
            # so we check the dtype explicitly
            assert res2.axes[0].labels.dtype == arr2.axes[0].labels.dtype
            assert res2.axes[1].labels.dtype == arr2.axes[1].labels.dtype

    def test_repr(self):
        with open_excel(visible=False) as wb:
            sheet = wb[0]
            assert re.match(r'<larray.inout.xw_excel.Sheet \[Book\d+]Sheet1>', repr(sheet))


@needs_xlwings
class TestRange:
    def test_scalar_convert(self):
        with open_excel(visible=False) as wb:
            sheet = wb[0]
            # set a few values
            sheet['A1'] = 1
            rng = sheet['A1']
            assert int(rng) == 1
            assert float(rng) == 1.0
            assert rng.__index__() == 1

            sheet['A2'] = 1.0
            rng = sheet['A2']
            assert int(rng) == 1
            assert float(rng) == 1.0
            # Excel stores everything as float so we cannot really make the difference between 1 and 1.0
            assert rng.__index__() == 1

            sheet['A3'] = 1.5
            rng = sheet['A3']
            assert int(rng) == 1
            assert float(rng) == 1.5
            with pytest.raises(TypeError, match="only integer scalars can be converted to a scalar index"):
                rng.__index__()

    def test_asarray(self):
        with open_excel(visible=False) as wb:
            sheet = wb[0]

            arr1 = ndtest((2, 3))
            # no header so that we have an uniform dtype for the whole sheet
            sheet['A1'] = arr1
            res1 = np.asarray(sheet['A1:C2'])
            assert np.array_equal(res1, arr1.data)
            assert res1.dtype == arr1.dtype

            arr1 = ndtest([Axis(2), Axis(3)])
            # no header so that we have an uniform dtype for the whole sheet
            sheet['A1'] = arr1
            res1 = asarray(sheet['A1:C2'])
            assert res1.equals(arr1)
            assert res1.dtype == arr1.dtype

    # this tests Range.__getattr__ with an Array attribute
    def test_aggregate(self):
        with open_excel(visible=False) as wb:
            sheet = wb[0]

            arr1 = ndtest((2, 3))
            # no header so that we have an uniform dtype for the whole sheet
            sheet['A1'] = arr1
            res = sheet['A1:C2'].sum()
            assert res == 15

    def test_repr(self):
        with open_excel(visible=False) as wb:
            sheet = wb[0]

            arr1 = ndtest((2, 3))
            sheet['A1'] = arr1
            res = repr(sheet['A1:C2'])
            assert res == """\
{0}*\\{1}*  0  1  2
        0  0  1  2
        1  3  4  5"""


# ================ #
# Test ExcelReport #
# ================ #


@needs_xlwings
def test_excel_report_init():
    # No argument
    ExcelReport()
    # with template dir
    ExcelReport(EXAMPLE_EXCEL_TEMPLATES_DIR)
    # with graphs_per_row
    ExcelReport(graphs_per_row=2)


@needs_xlwings
def test_excel_report_setting_template():
    excel_report = ExcelReport()

    # test setting template dir
    # 1) wrong template dir
    wrong_template_dir = r"C:\Wrong\Directory\Path"
    msg = f"The directory {wrong_template_dir} could not be found"
    with pytest.raises(ValueError, match=re.escape(msg)):
        excel_report.template_dir = wrong_template_dir
    # 2) correct path
    excel_report.template_dir = EXAMPLE_EXCEL_TEMPLATES_DIR
    assert excel_report.template_dir == EXAMPLE_EXCEL_TEMPLATES_DIR

    # test setting template file
    # 1) wrong extension
    template_file = 'wrong_extension.txt'
    msg = "Extension for the excel template file must be '.crtx' instead of .txt"
    with pytest.raises(ValueError, match=re.escape(msg)):
        excel_report.template = template_file
    # 2) add .crtx extension if no extension
    template_name = 'Line'
    excel_report.template = template_name
    assert excel_report.template == os.path.join(EXAMPLE_EXCEL_TEMPLATES_DIR, template_name + '.crtx')


@needs_xlwings
def test_excel_report_sheets():
    report = ExcelReport()
    # test adding new sheets
    report.new_sheet('Population')
    report.new_sheet('Births')
    report.new_sheet('Deaths')
    # test warning if sheet already exists
    with must_warn(UserWarning, msg="Sheet 'Population' already exists in the report and will be reset"):
        sheet_population2 = report.new_sheet('Population')  # noqa: F841
    # test sheet_names()
    assert report.sheet_names() == ['Population', 'Births', 'Deaths']


@needs_xlwings
def test_excel_report_titles():
    excel_report = ExcelReport()

    # test dumping titles
    sheet_titles = excel_report.new_sheet('Titles')
    # 1) default
    sheet_titles.add_title('Default title')
    # 2) specify width and height
    width, height = 1100, 100
    sheet_titles.add_title(f'Width = {width} and Height = {height}',
                           width=width, height=height)
    # 3) specify fontsize
    fontsize = 13
    sheet_titles.add_title(f'Fontsize = {fontsize}', fontsize=fontsize)

    # generate Excel file
    fpath = 'test_excel_report_titles.xlsx'
    excel_report.to_excel(fpath)


@needs_xlwings
@needs_pytables
def test_excel_report_arrays():
    excel_report = ExcelReport(EXAMPLE_EXCEL_TEMPLATES_DIR)
    demo = load_example_data('demography_eurostat')
    population = demo.population
    population_be = population['Belgium']
    population_be_nan = population_be.astype(float)
    population_be_nan[2013] = nan

    # fake simple user defined function to test if any user defined function (and their arguments)
    # is actually called and correctly handled.
    def customize_func(obj_chart, alternative_title):
        obj_chart.HasTitle = True
        obj_chart.ChartTitle.Caption = alternative_title

    # test dumping arrays
    sheet_graphs = excel_report.new_sheet('Graphs')
    # 1) default
    sheet_graphs.add_title('No template')
    sheet_graphs.add_graph(population_be['Female'], 'Pop Belgium Female')
    sheet_graphs.add_graph(population_be, 'Pop Belgium')
    sheet_graphs.add_graph(population_be_nan, 'Pop Belgium with nans')
    # 2) no title
    sheet_graphs.add_title('No title graph')
    sheet_graphs.add_graph(population_be)
    # 3) specify width and height
    sheet_graphs.add_title('Alternative Width and Height')
    width, height = 500, 300
    sheet_graphs.add_graph(population_be, f'width = {width} and Height = {height}',
                           width=width, height=height)
    # 4) specify template
    template_name = 'Line_Marker'
    sheet_graphs.add_title(f'Template = {template_name}')
    sheet_graphs.add_graph(population_be, f'Template = {template_name}', template_name)
    # 5) specify min_y and max_y
    sheet_graphs.add_graph(population_be, 'min (max) Y axis is 5e6 (6e6)', min_y=5e6, max_y=6e6)
    # 6) specify the space between 2 ticks
    sheet_graphs.add_graph(population_be, 'periodicity = every 2 years', xticks_spacing=2)
    # 7) pass a user defined function
    sheet_graphs.add_graph(population_be, customize_func=customize_func,
                           customize_kwargs={'alternative_title': 'alternative title'})

    # test setting default size
    # 1) pass a not registered kind of item
    item_type = 'unknown_item'
    msg = f"Item type {item_type} is not registered. Please choose in list ['graph', 'title']"
    with pytest.raises(ValueError, match=re.escape(msg)):
        sheet_graphs.set_item_default_size(item_type, width, height)
    # 2) update default size for graphs
    sheet_graphs.set_item_default_size('graph', width, height)
    sheet_graphs.add_title('Using Defined Sizes For All Graphs')
    sheet_graphs.add_graph(population_be, 'Pop Belgium')

    # test setting default number of graphs per row
    sheet_graphs = excel_report.new_sheet('Graphs2')
    sheet_graphs.graphs_per_row = 2
    sheet_graphs.add_title('graphs_per_row = 2')
    for combined_labels, subset in population.items(('country', 'gender')):
        title = ' - '.join([str(label) for label in combined_labels])
        sheet_graphs.add_graph(subset, title)

    # testing add_graphs
    sheet_graphs = excel_report.new_sheet('Graphs3')
    # 1) default
    sheet_graphs.add_title('add_graphs')
    sheet_graphs.add_graphs({'Population for {country} - {gender}': population},
                            {'gender': population.gender, 'country': population.country},
                            template='line', width=350, height=250, graphs_per_row=3)
    # 2) specify min_y and max_y
    sheet_graphs.add_title('min_y = 0 and max_y = 50e6')
    sheet_graphs.add_graphs({'Population for {country} - {gender}': population},
                            {'gender': population.gender, 'country': population.country},
                            template='line', width=350, height=250, graphs_per_row=3,
                            min_y=0, max_y=50e6)
    # 3) specify the space between 2 ticks
    sheet_graphs.add_title('periodicity = every 2 years')
    sheet_graphs.add_graphs({'Population for {country} - {gender}': population},
                            {'gender': population.gender, 'country': population.country},
                            template='line', width=350, height=250, graphs_per_row=3,
                            xticks_spacing=2)
    # 4) pass a user defined function
    sheet_graphs.add_title('user defined function')
    sheet_graphs.add_graphs({'Population for {country} - {gender}': population},
                            {'gender': population.gender, 'country': population.country},
                            template='line', width=350, height=250, graphs_per_row=3,
                            customize_func=customize_func,
                            customize_kwargs={'alternative_title': 'alternative title'})

    # generate Excel file
    fpath = 'test_excel_report_arrays.xlsx'
    excel_report.to_excel(fpath)


if __name__ == "__main__":
    pytest.main()



import pandas as pd

from larray.inout.pandas import from_frame

__all__ = ['read_stata']


def read_stata(filepath_or_buffer, index_col=None, sort_rows=False, sort_columns=False, **kwargs):
    r"""
    Reads Stata .dta file and returns an Array with the contents

    Parameters
    ----------
    filepath_or_buffer : str or file-like object
        Path to .dta file or a file handle.
    index_col : str or None, optional
        Name of column to set as index. Defaults to None.
    sort_rows : bool, optional
        Whether or not to sort the rows alphabetically (sorting is more efficient than not sorting).
        This only makes sense in combination with index_col. Defaults to False.
    sort_columns : bool, optional
        Whether or not to sort the columns alphabetically (sorting is more efficient than not sorting).
        Defaults to False.

    Returns
    -------
    Array

    See Also
    --------
    Array.to_stata

    Notes
    -----
    The round trip to Stata (Array.to_stata followed by read_stata) loose the name of the "column" axis.

    Examples
    --------
    >>> read_stata('test.dta')                   # doctest: +SKIP
    {0}\{1}  row  country  sex
          0    0       BE    F
          1    1       FR    M
          2    2       FR    F
    >>> read_stata('test.dta', index_col='row')  # doctest: +SKIP
    row\{1}  country  sex
          0       BE    F
          1       FR    M
          2       FR    F
    """
    df = pd.read_stata(filepath_or_buffer, index_col=index_col, **kwargs)
    return from_frame(df, sort_rows=sort_rows, sort_columns=sort_columns)



import os
import re
import sys
import inspect
from contextlib import contextmanager

import pytest
import datetime as dt
import numpy as np
import pandas as pd
try:
    import xlwings as xw
except ImportError:
    xw = None
try:
    import tables
except ImportError:
    tables = None
try:
    import openpyxl
except ImportError:
    openpyxl = None
try:
    import xlsxwriter
except ImportError:
    xlsxwriter = None

from larray import Array, isnan, asarray, Metadata


SKIP_EXCEL_TESTS = False
TESTDATADIR = os.path.dirname(__file__)


def inputpath(relpath):
    """
    Parameters
    ----------
    relpath: str
        path relative to current module

    Returns
    -------
    absolute path to input data file
    """
    return os.path.join(TESTDATADIR, 'data', relpath)


# XXX: maybe we should force value groups to use tuple and families (group of groups to use lists, or vice versa, so
# that we know which is which) or use a class, just for that? group(a, b, c) vs family(group(a), b, c)


def assert_equal_factory(test_func):
    def assert_equal(a, b):
        if isinstance(a, Array) and isinstance(b, Array) and a.axes != b.axes:
            raise AssertionError(f"axes differ:\n{a.axes.info}\n\nvs\n\n{b.axes.info}")
        if not isinstance(a, (np.ndarray, Array)):
            a = np.asarray(a)
        if not isinstance(b, (np.ndarray, Array)):
            b = np.asarray(b)
        if a.shape != b.shape:
            raise AssertionError(f"shapes differ: {a.shape} != {b.shape}")
        equal = test_func(a, b)
        if not equal.all():
            # XXX: for some reason ndarray[bool_larray] does not work as we would like, so we cannot do b[~equal]
            #      directly. I should at least understand why this happens and fix this if possible.
            notequal = np.asarray(~equal)
            raise AssertionError(f"\ngot:\n\n{a[notequal]}\n\nexpected:\n\n{b[notequal]}")
    return assert_equal


def assert_larray_equal_factory(test_func, convert=True, check_axes=False):
    def assert_equal(a, b):
        if convert:
            a = asarray(a)
            b = asarray(b)
        if check_axes and a.axes != b.axes:
            raise AssertionError(f"axes differ:\n{a.axes.info}\n\nvs\n\n{b.axes.info}")
        equal = test_func(a, b)
        if not equal.all():
            notequal = ~equal
            raise AssertionError(f"\ngot:\n\n{a[notequal]}\n\nexpected:\n\n{b[notequal]}")
    return assert_equal


def assert_nparray_equal_factory(test_func, convert=True, check_shape=False):
    def assert_equal(a, b):
        if convert:
            a = np.asarray(a)
            b = np.asarray(b)
        if check_shape and a.shape != b.shape:
            raise AssertionError(f"shapes differ: {a.shape} != {b.shape}")
        equal = test_func(a, b)
        if not equal.all():
            notequal = ~equal
            raise AssertionError(f"\ngot:\n\n{a[notequal]}\n\nexpected:\n\n{b[notequal]}")
    return assert_equal


def equal(a, b):
    return a == b


def nan_equal(a, b):
    return (a == b) | (isnan(a) & isnan(b))


# numpy.testing.assert_array_equal/assert_equal would work too but it does not (as of numpy 1.10) display specifically
# the non equal items
# TODO: this is defined for backward compatibility only (until we update all tests to use either assert_larray* or
#       assert_nparray*)
assert_array_equal = assert_equal_factory(equal)
assert_array_nan_equal = assert_equal_factory(nan_equal)

assert_larray_equal = assert_larray_equal_factory(equal, check_axes=True)
assert_larray_nan_equal = assert_larray_equal_factory(nan_equal, check_axes=True)

assert_larray_equiv = assert_larray_equal_factory(equal)
assert_larray_nan_equiv = assert_larray_equal_factory(nan_equal)

assert_nparray_equal = assert_nparray_equal_factory(equal, check_shape=True)
assert_nparray_nan_equal = assert_nparray_equal_factory(nan_equal, check_shape=True)

assert_nparray_equiv = assert_nparray_equal_factory(equal)
assert_nparray_nan_equiv = assert_nparray_equal_factory(nan_equal)


def assert_axis_eq(axis1, axis2):
    assert axis1.equals(axis2)


def tmp_path(tmpdir, fname):
    return os.path.join(tmpdir.strpath, fname)


@pytest.fixture
def meta():
    title = 'test array'
    description = 'Array used for testing'
    author = 'John Cleese'
    location = 'Ministry of Silly Walks'
    office_number = 42
    score = 9.70
    date = pd.Timestamp(dt.datetime(1970, 3, 21))
    return Metadata([('title', title), ('description', description), ('author', author),
                     ('location', location), ('office_number', office_number),
                     ('score', score), ('date', date)])


needs_pytables = pytest.mark.skipif(tables is None, reason="pytables is required for this test")
needs_xlwings = pytest.mark.skipif(SKIP_EXCEL_TESTS or xw is None, reason="xlwings is required for this test")
needs_openpyxl = pytest.mark.skipif(SKIP_EXCEL_TESTS or openpyxl is None, reason="openpyxl is required for this test")
needs_xlsxwriter = pytest.mark.skipif(SKIP_EXCEL_TESTS or xlsxwriter is None,
                                      reason="xlsxwriter is required for this test")

needs_python37 = pytest.mark.skipif(sys.version_info < (3, 7), reason="Python 3.7 is required for this test")


@contextmanager
def must_warn(warn_cls=None, msg=None, match=None, check_file=True, check_num=True):
    if msg is not None and match is not None:
        raise ValueError("bad test: can't use both msg and match arguments")
    elif msg is not None:
        match = re.escape(msg)

    try:
        with pytest.warns(warn_cls, match=match) as caught_warnings:
            yield caught_warnings
    finally:
        if check_num:
            assert len(caught_warnings) == 1
        if check_file:
            caller_path = inspect.stack()[2].filename
            warning_path = caught_warnings[0].filename
            assert warning_path == caller_path, \
                f"{warning_path} != {caller_path}"


@contextmanager
def must_raise(warn_cls=None, msg=None, match=None):
    if msg is not None and match is not None:
        raise ValueError("bad test: can't use both msg and match arguments")
    elif msg is not None:
        match = re.escape(msg)

    try:
        with pytest.raises(warn_cls, match=match) as error:
            yield error
    finally:
        pass



class ExprNode:
    # method factory
    def _binop(opname):
        def opmethod(self, other):
            return BinaryOp(opname, self, other)

        opmethod.__name__ = f'__{opname}__'
        return opmethod

    __matmul__ = _binop('matmul')
    __ror__ = _binop('ror')
    __or__ = _binop('or')
    __rxor__ = _binop('rxor')
    __xor__ = _binop('xor')
    __rand__ = _binop('rand')
    __and__ = _binop('and')
    __rrshift__ = _binop('rrshift')
    __rshift__ = _binop('rshift')
    __rlshift__ = _binop('rlshift')
    __lshift__ = _binop('lshift')
    __rpow__ = _binop('rpow')
    __pow__ = _binop('pow')
    __rdivmod__ = _binop('rdivmod')
    __divmod__ = _binop('divmod')
    __rmod__ = _binop('rmod')
    __mod__ = _binop('mod')
    # div and rdiv are not longer used on Python3+
    __rfloordiv__ = _binop('rfloordiv')
    __floordiv__ = _binop('floordiv')
    __rtruediv__ = _binop('rtruediv')
    __truediv__ = _binop('truediv')
    __rmul__ = _binop('rmul')
    __mul__ = _binop('mul')
    __rsub__ = _binop('rsub')
    __sub__ = _binop('sub')
    __radd__ = _binop('radd')
    __add__ = _binop('add')
    __ge__ = _binop('ge')
    __gt__ = _binop('gt')
    __ne__ = _binop('ne')
    __eq__ = _binop('eq')
    __le__ = _binop('le')
    __lt__ = _binop('lt')

    def _unaryop(opname):
        def opmethod(self):
            return UnaryOp(opname, self)

        opmethod.__name__ = f'__{opname}__'
        return opmethod

    # unary ops do not need broadcasting so do not need to be overridden
    __neg__ = _unaryop('neg')
    __pos__ = _unaryop('pos')
    __abs__ = _unaryop('abs')
    __invert__ = _unaryop('invert')

    def evaluate(self, context):
        raise NotImplementedError()


def expr_eval(expr, context):
    # in the end it all comes down to AxisReference.evaluate(AxisCollection) which returns the Axis
    return expr.evaluate(context) if isinstance(expr, ExprNode) else expr


class BinaryOp(ExprNode):
    def __init__(self, op, expr1, expr2):
        self.opname = f'__{op}__'
        self.expr1 = expr1
        self.expr2 = expr2

    def evaluate(self, context):
        # TODO: implement eval via numexpr
        expr1 = expr_eval(self.expr1, context)
        expr2 = expr_eval(self.expr2, context)
        return getattr(expr1, self.opname)(expr2)


class UnaryOp(ExprNode):
    def __init__(self, op, expr):
        self.opname = f'__{op}__'
        self.expr = expr

    def evaluate(self, context):
        # TODO: implement eval via numexpr
        expr = expr_eval(self.expr, context)
        return getattr(expr, self.opname)()



import warnings
import sys
import os

import numpy as np
import pandas as pd
try:
    import xlwings as xw
except ImportError:
    xw = None
try:
    import xlsxwriter
except ImportError:
    xlsxwriter = None

from larray.core.array import Array, asarray
from larray.core.constants import nan
from larray.core.group import _translate_sheet_name
from larray.core.metadata import Metadata
from larray.util.misc import deprecate_kwarg
from larray.inout.session import register_file_handler
from larray.inout.common import _get_index_col, FileHandler
from larray.inout.pandas import df_asarray
from larray.inout.xw_excel import open_excel
from larray.example import get_example_filepath             # noqa: F401


__all__ = ['read_excel']


@deprecate_kwarg('nb_index', 'nb_axes', arg_converter=lambda x: x + 1)
@deprecate_kwarg('sheetname', 'sheet')
# We use "# doctest: +SKIP" for all tests because they work only if openpyxl (an *optional* dependency) is installed
def read_excel(filepath, sheet=0, nb_axes=None, index_col=None, fill_value=nan, na=nan,
               sort_rows=False, sort_columns=False, wide=True, engine=None, range=slice(None), **kwargs):
    r"""
    Reads excel file from sheet name and returns an Array with the contents

    Parameters
    ----------
    filepath : str
        Path where the Excel file has to be read or use -1 to refer to the currently active workbook.
    sheet : str, Group or int, optional
        Name or index of the Excel sheet containing the array to be read.
        By default the array is read from the first sheet.
    nb_axes : int, optional
        Number of axes of output array. The first ``nb_axes`` - 1 columns and the header of the Excel sheet will be used
        to set the axes of the output array. If not specified, the number of axes is given by the position of the
        first column header including a ``\`` character plus one. If no column header includes a ``\`` character, the
        array is assumed to have one axis. Defaults to None.
    index_col : list, optional
        Positions of columns for the n-1 first axes (ex. [0, 1, 2, 3]). Defaults to None (see nb_axes above).
    fill_value : scalar or Array, optional
        Value used to fill cells corresponding to label combinations which are not present in the input.
        Defaults to NaN.
    sort_rows : bool, optional
        Whether or not to sort the rows alphabetically (sorting is more efficient than not sorting). Defaults to False.
    sort_columns : bool, optional
        Whether or not to sort the columns alphabetically (sorting is more efficient than not sorting).
        Defaults to False.
    wide : bool, optional
        Whether or not to assume the array is stored in "wide" format.
        If False, the array is assumed to be stored in "narrow" format: one column per axis plus one value column.
        Defaults to True.
    engine : {'xlwings', 'openpyxl', 'xlrd'}, optional
        Engine to use to read the Excel file.
        The 'xlrd' engine must be used to read Excel files with the old '.xls' extension.
        Either 'xlwings' or 'openpyxl' can be used to read Excel files with the standard '.xlsx' extension.
        Defaults to 'xlwings' if the module is installed, 'openpyxl' otherwise.
    range : str, optional
        Range to load the array from (only supported for the 'xlwings' engine). Defaults to slice(None) which loads
        the whole sheet, ignoring blank cells in the bottom right corner.
    **kwargs

    Returns
    -------
    Array

    Examples
    --------
    >>> fname = get_example_filepath('examples.xlsx')

    Read array from first sheet

    >>> # The data below is derived from a subset of the demo_pjan table from Eurostat
    >>> read_excel(fname)                                                               # doctest: +SKIP
    country  gender\time      2013      2014      2015
    Belgium         Male   5472856   5493792   5524068
    Belgium       Female   5665118   5687048   5713206
     France         Male  31772665  32045129  32174258
     France       Female  33827685  34120851  34283895
    Germany         Male  39380976  39556923  39835457
    Germany       Female  41142770  41210540  41362080

    Read array from a specific sheet

    >>> # The data below is derived from a subset of the demo_fasec table from Eurostat
    >>> read_excel(fname, 'births')                                                     # doctest: +SKIP
    country  gender\time    2013    2014    2015
    Belgium         Male   64371   64173   62561
    Belgium       Female   61235   60841   59713
     France         Male  415762  418721  409145
     France       Female  396581  400607  390526
    Germany         Male  349820  366835  378478
    Germany       Female  332249  348092  359097

    Missing label combinations

    Let us take a look inside the sheet 'population_missing_values'. Note the missing label combinations:
    (Paris, male) and (New York, female): ::

        country  gender\time      2013      2014      2015
        Belgium         Male   5472856   5493792   5524068
        Belgium       Female   5665118   5687048   5713206
         France       Female  33827685  34120851  34283895
        Germany         Male  39380976  39556923  39835457

    By default, cells associated with missing label combinations are filled with NaN. In that case, an int array
    is converted to a float array.

    >>> read_excel(fname, sheet='population_missing_values')                            # doctest: +SKIP
    country  gender\time        2013        2014        2015
    Belgium         Male   5472856.0   5493792.0   5524068.0
    Belgium       Female   5665118.0   5687048.0   5713206.0
     France         Male         nan         nan         nan
     France       Female  33827685.0  34120851.0  34283895.0
    Germany         Male  39380976.0  39556923.0  39835457.0
    Germany       Female         nan         nan         nan

    Using the ``fill_value`` argument, you can choose another value to use to fill missing cells.

    >>> read_excel(fname, sheet='population_missing_values', fill_value=0)              # doctest: +SKIP
    country  gender\time      2013      2014      2015
    Belgium         Male   5472856   5493792   5524068
    Belgium       Female   5665118   5687048   5713206
     France         Male         0         0         0
     France       Female  33827685  34120851  34283895
    Germany         Male  39380976  39556923  39835457
    Germany       Female         0         0         0

    Specify the number of axes of the output array (useful when the name of the last axis is implicit)

    The content of the sheet 'missing_axis_name' is: ::

        country  gender      2013      2014      2015
        Belgium    Male   5472856   5493792   5524068
        Belgium  Female   5665118   5687048   5713206
         France    Male  31772665  32045129  32174258
         France  Female  33827685  34120851  34283895
        Germany    Male  39380976  39556923  39835457
        Germany  Female  41142770  41210540  41362080

    >>> # read the array stored in the sheet 'population_missing_axis_name' as is
    >>> arr = read_excel(fname, sheet='population_missing_axis_name')                   # doctest: +SKIP
    >>> # we expected a 3 x 2 x 3 array with data of type int
    >>> # but we got a 6 x 4 array with data of type object
    >>> arr.info                                                                        # doctest: +SKIP
    6 x 4
     country [6]: 'Belgium' 'Belgium' 'France' 'France' 'Germany' 'Germany'
     {1} [4]: 'gender' '2013' '2014' '2015'
    dtype: object
    memory used: 192 bytes
    >>> # using argument 'nb_axes', you can force the number of axes of the output array
    >>> arr = read_excel(fname, sheet='population_missing_axis_name', nb_axes=3)        # doctest: +SKIP
    >>> # as expected, we have a 3 x 2 x 3 array with data of type int
    >>> arr.info                                                                        # doctest: +SKIP
    3 x 2 x 3
     country [3]: 'Belgium' 'France' 'Germany'
     gender [2]: 'Male' 'Female'
     {2} [3]: 2013 2014 2015
    dtype: int64
    memory used: 144 bytes

    Read array saved in "narrow" format (wide=False)

    Let us take a look inside the sheet 'population_narrow' where the data is stored in a 'narrow' format: ::

        country  time     value
        Belgium  2013  11137974
        Belgium  2014  11180840
        Belgium  2015  11237274
         France  2013  65600350
         France  2014  66165980
         France  2015  66458153

    >>> # to read arrays stored in 'narrow' format, you must pass wide=False to read_excel
    >>> read_excel(fname, 'population_narrow_format', wide=False)                       # doctest: +SKIP
    country\time      2013      2014      2015
         Belgium  11137974  11180840  11237274
          France  65600350  66165980  66458153

    Extract array from a given range (xlwings only)

    >>> read_excel(fname, 'population_births_deaths', range='A9:E15')                   # doctest: +SKIP
    country  gender\time    2013    2014    2015
    Belgium         Male   64371   64173   62561
    Belgium       Female   61235   60841   59713
     France         Male  415762  418721  409145
     France       Female  396581  400607  390526
    Germany         Male  349820  366835  378478
    Germany       Female  332249  348092  359097
    """
    if not np.isnan(na):
        fill_value = na
        warnings.warn("read_excel `na` argument has been renamed to `fill_value`. Please use that instead.",
                      FutureWarning, stacklevel=2)

    sheet = _translate_sheet_name(sheet)

    if engine is None:
        engine = 'xlwings' if xw is not None else 'openpyxl'

    index_col = _get_index_col(nb_axes, index_col, wide)

    if engine == 'xlwings':
        if kwargs:
            first_kwarg = list(kwargs.keys())[0]
            raise TypeError(f"'{first_kwarg}' is an invalid keyword argument for this function when using the xlwings "
                            f"backend")
        from larray.inout.xw_excel import open_excel
        with open_excel(filepath) as wb:
            return wb[sheet][range].load(index_col=index_col, fill_value=fill_value, sort_rows=sort_rows,
                                         sort_columns=sort_columns, wide=wide)
    else:
        # TODO: add support for range argument (using usecols, skiprows and nrows arguments of pandas.read_excel)
        df = pd.read_excel(filepath, sheet, index_col=index_col, engine=engine, **kwargs)
        return df_asarray(df, sort_rows=sort_rows, sort_columns=sort_columns, raw=index_col is None,
                          fill_value=fill_value, wide=wide)


@register_file_handler('pandas_excel', ['xls', 'xlsx'] if xw is None else None)
class PandasExcelHandler(FileHandler):
    r"""
    Handler for Excel files using Pandas.
    """
    def __init__(self, fname, overwrite_file=False):
        super(PandasExcelHandler, self).__init__(fname, overwrite_file)

    def _open_for_read(self):
        engine = 'openpyxl' if sys.version_info < (3, 7) else None
        self.handle = pd.ExcelFile(self.fname, engine=engine)

    def _open_for_write(self):
        _, ext = os.path.splitext(self.fname)
        engine = 'xlsxwriter' if ext == '.xlsx' and xlsxwriter is not None else None
        self.handle = pd.ExcelWriter(self.fname, engine=engine)

    def list_items(self):
        return [(name, 'Array') for name in self.handle.sheet_names if name != '__metadata__']

    def _read_item(self, key, type, *args, **kwargs):
        if type == 'Array':
            df = self.handle.parse(key, *args, **kwargs)
            return df_asarray(df, raw=True)
        else:
            raise TypeError()

    def _dump_item(self, key, value, *args, **kwargs):
        kwargs['engine'] = self.handle.engine
        if isinstance(value, Array):
            value.to_excel(self.handle, key, *args, **kwargs)
        else:
            raise TypeError()

    def _read_metadata(self):
        sheet_meta = '__metadata__'
        if sheet_meta in self.handle.sheet_names:
            meta = read_excel(self.handle, sheet_meta, engine=self.handle.engine, wide=False)
            return Metadata.from_array(meta)
        else:
            return Metadata()

    def _dump_metadata(self, metadata):
        if len(metadata) > 0:
            metadata = asarray(metadata)
            metadata.to_excel(self.handle, '__metadata__', engine=self.handle.engine, wide=False, value_name='')

    def save(self):
        pass

    def close(self):
        self.handle.close()


@register_file_handler('xlwings_excel', ['xls', 'xlsx'] if xw is not None else None)
class XLWingsHandler(FileHandler):
    r"""
    Handler for Excel files using XLWings.
    """
    def __init__(self, fname, overwrite_file=False):
        super(XLWingsHandler, self).__init__(fname, overwrite_file)

    def _get_original_file_name(self):
        # for XLWingsHandler, no need to create a temporary file, the job is already done in the Workbook class
        pass

    def _open_for_read(self):
        self.handle = open_excel(self.fname)

    def _open_for_write(self):
        self.handle = open_excel(self.fname, overwrite_file=self.overwrite_file)

    def list_items(self):
        return [(name, 'Array') for name in self.handle.sheet_names() if name != '__metadata__']

    def _read_item(self, key, type, *args, **kwargs):
        if type == 'Array':
            return self.handle[key].load(*args, **kwargs)
        else:
            raise TypeError()

    def _dump_item(self, key, value, *args, **kwargs):
        if isinstance(value, Array):
            self.handle[key] = value.dump(*args, **kwargs)
        else:
            raise TypeError()

    def _read_metadata(self):
        sheet_meta = '__metadata__'
        if sheet_meta in self.handle:
            meta = self.handle[sheet_meta].load(wide=False)
            return Metadata.from_array(meta)
        else:
            return Metadata()

    def _dump_metadata(self, metadata):
        if len(metadata) > 0:
            metadata = asarray(metadata)
            self.handle['__metadata__'] = metadata.dump(wide=False, value_name='')

    def save(self):
        self.handle.save()

    def close(self):
        self.handle.close()



from itertools import product

import numpy as np
import pandas as pd

from larray.core.array import Array
from larray.core.axis import Axis, AxisCollection
from larray.core.constants import nan
from larray.util.misc import unique_list


def decode(s, encoding='utf-8', errors='strict'):
    if isinstance(s, bytes):
        return s.decode(encoding, errors)
    else:
        assert s is None or isinstance(s, str), "unexpected " + str(type(s))
        return s


def parse(s):
    r"""
    Used to parse the "folded" axis ticks (usually periods).
    """
    # parameters can be strings or numbers
    if isinstance(s, str):
        s = s.strip()
        low = s.lower()
        if low == 'true':
            return True
        elif low == 'false':
            return False
        elif s.isdigit():
            return int(s)
        else:
            try:
                return float(s)
            except ValueError:
                return s
    else:
        return s


def index_to_labels(idx, sort=True):
    r"""
    Returns unique labels for each dimension.
    """
    if isinstance(idx, pd.MultiIndex):
        if sort:
            return list(idx.levels)
        else:
            return [unique_list(idx.get_level_values(label)) for label in range(idx.nlevels)]
    else:
        assert isinstance(idx, pd.Index)
        labels = list(idx.values)
        return [sorted(labels) if sort else labels]


def cartesian_product_df(df, sort_rows=False, sort_columns=False, fill_value=nan, **kwargs):
    idx = df.index
    labels = index_to_labels(idx, sort=sort_rows)
    if isinstance(idx, pd.MultiIndex):
        if sort_rows:
            new_index = pd.MultiIndex.from_product(labels)
        else:
            new_index = pd.MultiIndex.from_tuples(list(product(*labels)))
    else:
        if sort_rows:
            new_index = pd.Index(labels[0], name=idx.name)
        else:
            new_index = idx
    columns = sorted(df.columns) if sort_columns else list(df.columns)
    # the prodlen test is meant to avoid the more expensive array_equal test
    prodlen = np.prod([len(axis_labels) for axis_labels in labels])
    if prodlen == len(df) and columns == list(df.columns) and np.array_equal(idx.values, new_index.values):
        return df, labels
    return df.reindex(index=new_index, columns=columns, fill_value=fill_value, **kwargs), labels


def from_series(s, sort_rows=False, fill_value=nan, meta=None, **kwargs):
    r"""
    Converts Pandas Series into Array.

    Parameters
    ----------
    s : Pandas Series
        Input Pandas Series.
    sort_rows : bool, optional
        Whether or not to sort the rows alphabetically. Defaults to False.
    fill_value : scalar, optional
        Value used to fill cells corresponding to label combinations which are not present in the input Series.
        Defaults to NaN.
    meta : list of pairs or dict or OrderedDict or Metadata, optional
        Metadata (title, description, author, creation_date, ...) associated with the array.
        Keys must be strings. Values must be of type string, int, float, date, time or datetime.

    Returns
    -------
    Array

    See Also
    --------
    Array.to_series

    Examples
    --------
    >>> from larray import ndtest
    >>> s = ndtest((2, 2, 2), dtype=float).to_series()
    >>> s                                                                             # doctest: +NORMALIZE_WHITESPACE
    a   b   c
    a0  b0  c0    0.0
            c1    1.0
        b1  c0    2.0
            c1    3.0
    a1  b0  c0    4.0
            c1    5.0
        b1  c0    6.0
            c1    7.0
    dtype: float64
    >>> from_series(s)
     a  b\c   c0   c1
    a0   b0  0.0  1.0
    a0   b1  2.0  3.0
    a1   b0  4.0  5.0
    a1   b1  6.0  7.0
    """
    if isinstance(s.index, pd.MultiIndex):
        # TODO: use argument sort=False when it will be available
        # (see https://github.com/pandas-dev/pandas/issues/15105)
        df = s.unstack(level=-1, fill_value=fill_value)
        # pandas (un)stack and pivot(_table) methods return a Dataframe/Series with sorted index and columns
        if not sort_rows:
            labels = index_to_labels(s.index, sort=False)
            if isinstance(df.index, pd.MultiIndex):
                index = pd.MultiIndex.from_tuples(list(product(*labels[:-1])), names=s.index.names[:-1])
            else:
                index = labels[0]
            columns = labels[-1]
            df = df.reindex(index=index, columns=columns, fill_value=fill_value)
        return from_frame(df, sort_rows=sort_rows, sort_columns=sort_rows, fill_value=fill_value, meta=meta, **kwargs)
    else:
        name = decode(s.name, 'utf8') if s.name is not None else decode(s.index.name, 'utf8')
        if sort_rows:
            s = s.sort_index()
        return Array(s.values, Axis(s.index.values, name), meta=meta)


def from_frame(df, sort_rows=False, sort_columns=False, parse_header=False, unfold_last_axis_name=False,
               fill_value=nan, meta=None, cartesian_prod=True, **kwargs):
    r"""
    Converts Pandas DataFrame into Array.

    Parameters
    ----------
    df : pandas.DataFrame
        Input dataframe. By default, name and labels of the last axis are defined by the name and labels of the
        columns Index of the dataframe unless argument unfold_last_axis_name is set to True.
    sort_rows : bool, optional
        Whether or not to sort the rows alphabetically (sorting is more efficient than not sorting).
        Must be False if `cartesian_prod` is set to True.
        Defaults to False.
    sort_columns : bool, optional
        Whether or not to sort the columns alphabetically (sorting is more efficient than not sorting).
        Must be False if `cartesian_prod` is set to True.
        Defaults to False.
    parse_header : bool, optional
        Whether or not to parse columns labels. Pandas treats column labels as strings.
        If True, column labels are converted into int, float or boolean when possible. Defaults to False.
    unfold_last_axis_name : bool, optional
        Whether or not to extract the names of the last two axes by splitting the name of the last index column of the
        dataframe using ``\``. Defaults to False.
    fill_value : scalar, optional
        Value used to fill cells corresponding to label combinations which are not present in the input DataFrame.
        Defaults to NaN.
    meta : list of pairs or dict or OrderedDict or Metadata, optional
        Metadata (title, description, author, creation_date, ...) associated with the array.
        Keys must be strings. Values must be of type string, int, float, date, time or datetime.
    cartesian_prod : bool, optional
        Whether or not to expand the dataframe to a cartesian product dataframe as needed by Array.
        This is an expensive operation but is absolutely required if you cannot guarantee your dataframe is already
        well formed. If True, arguments `sort_rows` and `sort_columns` must be set to False.
        Defaults to True.

    Returns
    -------
    Array

    See Also
    --------
    Array.to_frame

    Examples
    --------
    >>> from larray import ndtest
    >>> df = ndtest((2, 2, 2)).to_frame()
    >>> df                                                                             # doctest: +NORMALIZE_WHITESPACE
    c      c0  c1
    a  b
    a0 b0   0   1
       b1   2   3
    a1 b0   4   5
       b1   6   7
    >>> from_frame(df)
     a  b\c  c0  c1
    a0   b0   0   1
    a0   b1   2   3
    a1   b0   4   5
    a1   b1   6   7

    Names of the last two axes written as ``before_last_axis_name\\last_axis_name``

    >>> df = ndtest((2, 2, 2)).to_frame(fold_last_axis_name=True)
    >>> df                                                                             # doctest: +NORMALIZE_WHITESPACE
            c0  c1
    a  b\c
    a0 b0    0   1
       b1    2   3
    a1 b0    4   5
       b1    6   7
    >>> from_frame(df, unfold_last_axis_name=True)
     a  b\c  c0  c1
    a0   b0   0   1
    a0   b1   2   3
    a1   b0   4   5
    a1   b1   6   7
    """
    axes_names = [decode(name, 'utf8') if isinstance(name, bytes) else name
                  for name in df.index.names]

    # handle 2 or more dimensions with the last axis name given using \
    if unfold_last_axis_name:
        if isinstance(axes_names[-1], str) and '\\' in axes_names[-1]:
            last_axes = [name.strip() for name in axes_names[-1].split('\\')]
            axes_names = axes_names[:-1] + last_axes
        else:
            axes_names += [None]
    else:
        axes_names += [df.columns.name]

    if cartesian_prod:
        df, axes_labels = cartesian_product_df(df, sort_rows=sort_rows, sort_columns=sort_columns,
                                               fill_value=fill_value, **kwargs)
    else:
        if sort_rows or sort_columns:
            raise ValueError('sort_rows and sort_columns cannot not be used when cartesian_prod is set to False. '
                             'Please call the method sort_axes on the returned array to sort rows or columns')
        axes_labels = index_to_labels(df.index, sort=False)

    # Pandas treats column labels as column names (strings) so we need to convert them to values
    last_axis_labels = [parse(cell) for cell in df.columns.values] if parse_header else list(df.columns.values)
    axes_labels.append(last_axis_labels)

    axes = AxisCollection([Axis(labels, name) for labels, name in zip(axes_labels, axes_names)])
    data = df.values.reshape(axes.shape)
    return Array(data, axes, meta=meta)


def set_dataframe_index_by_position(df, index_col_indices):
    """
    equivalent to Dataframe.set_index but with column indices, not column labels

    This is necessary to support creating an index from columns without a name or with duplicate names.

    Returns a new Dataframe
    """
    if not isinstance(index_col_indices, list):
        index_col_indices = [index_col_indices]
    index_col_indices_set = set(index_col_indices)
    index_col_values = [df.iloc[:, i] for i in index_col_indices]
    non_index_col_indices = [i for i in range(len(df.columns)) if i not in index_col_indices_set]
    # drop the index columns from the "normal" columns of the dataframe
    df = df.iloc[:, non_index_col_indices]
    # add them back as index columns
    df.set_index(index_col_values, inplace=True)
    return df


def df_asarray(df, sort_rows=False, sort_columns=False, raw=False, parse_header=True, wide=True, cartesian_prod=True,
               **kwargs):
    r"""
    Prepare Pandas DataFrame and then convert it into Array.

    Parameters
    ----------
    df : Pandas DataFrame
        Input dataframe.
    sort_rows : bool, optional
        Whether or not to sort the rows alphabetically (sorting is more efficient than not sorting).
        Must be False if `cartesian_prod` is set to True.
        Defaults to False.
    sort_columns : bool, optional
        Whether or not to sort the columns alphabetically (sorting is more efficient than not sorting).
        Must be False if `cartesian_prod` is set to True.
        Defaults to False.
    raw : bool, optional
        Whether or not to consider the input dataframe as a raw dataframe, i.e. read without index at all.
        If True, build the first N-1 axes of the output array from the first N-1 dataframe columns. Defaults to False.
    parse_header : bool, optional
        Whether or not to parse columns labels. Pandas treats column labels as strings.
        If True, column labels are converted into int, float or boolean when possible. Defaults to True.
    wide : bool, optional
        Whether or not to assume the array is stored in "wide" format.
        If False, the array is assumed to be stored in "narrow" format: one column per axis plus one value column.
        Defaults to True.
    cartesian_prod : bool, optional
        Whether or not to expand the dataframe to a cartesian product dataframe as needed by Array.
        This is an expensive operation but is absolutely required if you cannot guarantee your dataframe is already
        well formed. If True, arguments `sort_rows` and `sort_columns` must be set to False.
        Defaults to True.

    Returns
    -------
    Array
    """
    # we could inline df_asarray into the functions that use it, so that the original (non-cartesian) df is freed from
    # memory at this point, but it would be much uglier and would not lower the peak memory usage which happens during
    # cartesian_product_df.reindex

    # raw = True: the dataframe was read without index at all (ie 2D dataframe),
    # irrespective of the actual data dimensionality
    if raw:
        columns = df.columns.values.tolist()
        if wide:
            try:
                # take the first column which contains '\'
                pos_last = next(i for i, v in enumerate(columns) if isinstance(v, str) and '\\' in v)
            except StopIteration:
                # we assume first column will not contain data
                pos_last = 0

            # This is required to handle int column names (otherwise we can simply use column positions in set_index).
            # This is NOT the same as df.columns[list(range(...))] !
            df = set_dataframe_index_by_position(df, list(range(pos_last + 1)))
        else:
            df = set_dataframe_index_by_position(df, list(range(len(df.columns) - 1)))
            series = df.iloc[:, -1]
            series.name = df.index.name
            return from_series(series, sort_rows=sort_columns, **kwargs)

    # handle 1D arrays
    if len(df) == 1 and (pd.isnull(df.index.values[0])
                         or (isinstance(df.index.values[0], str) and df.index.values[0].strip() == '')):
        if parse_header:
            df.columns = pd.Index([parse(cell) for cell in df.columns.values], name=df.columns.name)
        series = df.iloc[0]
        series.name = df.index.name
        if sort_rows:
            raise ValueError('sort_rows=True is not valid for 1D arrays. Please use sort_columns instead.')
        res = from_series(series, sort_rows=sort_columns)
    else:
        def parse_axis_name(name):
            if isinstance(name, bytes):
                name = decode(name, 'utf8')
            if not name:
                name = None
            return name
        axes_names = [parse_axis_name(name) for name in df.index.names]
        unfold_last_axis_name = isinstance(axes_names[-1], str) and '\\' in axes_names[-1]
        res = from_frame(df, sort_rows=sort_rows, sort_columns=sort_columns, parse_header=parse_header,
                         unfold_last_axis_name=unfold_last_axis_name, cartesian_prod=cartesian_prod, **kwargs)

    # ugly hack to avoid anonymous axes converted as axes with name 'Unnamed: x' by pandas
    # we also take the opportunity to change axes with empty name to real anonymous axes (name is None) to
    # make them roundtrip correctly, based on the assumption that in an in-memory LArray an anonymouse axis is more
    # likely and useful than an Axis with an empty name.
    # TODO : find a more robust and elegant solution
    res = res.rename({axis: None for axis in res.axes if (isinstance(axis.name, str)
                                                          and (axis.name == '' or 'Unnamed:' in axis.name))})
    return res
